################################################################################
# Copyright (C) 2014 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


"""
Unit tests for `gate` module.
"""

import numpy as np

from bayespy.nodes import (Gate,
                           GaussianARD,
                           Gamma,
                           Categorical,
                           Bernoulli,
                           Multinomial)

from bayespy.inference.vmp.nodes.gaussian import GaussianMoments

from bayespy.utils import random
from bayespy.utils import linalg

from bayespy.utils.misc import TestCase


class TestGate(TestCase):
    """
    Unit tests for Gate node.
    """


    def test_init(self):
        """
        Test the creation of Gate node
        """

        # Gating scalar node
        Z = Categorical(np.ones(3)/3)
        X = GaussianARD(0, 1, shape=(), plates=(3,))
        Y = Gate(Z, X)
        self.assertEqual(Y.plates, ())
        self.assertEqual(Y.dims, ( (), () ))

        # Gating non-scalar node
        Z = Categorical(np.ones(3)/3)
        X = GaussianARD(0, 1, shape=(2,), plates=(3,))
        Y = Gate(Z, X)
        self.assertEqual(Y.plates, ())
        self.assertEqual(Y.dims, ( (2,), (2,2) ))

        # Plates from Z
        Z = Categorical(np.ones(3)/3, plates=(4,))
        X = GaussianARD(0, 1, shape=(2,), plates=(3,))
        Y = Gate(Z, X)
        self.assertEqual(Y.plates, (4,))
        self.assertEqual(Y.dims, ( (2,), (2,2) ))

        # Plates from X
        Z = Categorical(np.ones(3)/3)
        X = GaussianARD(0, 1, shape=(2,), plates=(4,3))
        Y = Gate(Z, X)
        self.assertEqual(Y.plates, (4,))
        self.assertEqual(Y.dims, ( (2,), (2,2) ))

        # Plates from Z and X
        Z = Categorical(np.ones(3)/3, plates=(5,))
        X = GaussianARD(0, 1, shape=(2,), plates=(4,1,3))
        Y = Gate(Z, X)
        self.assertEqual(Y.plates, (4,5))
        self.assertEqual(Y.dims, ( (2,), (2,2) ))

        # Gating non-default plate
        Z = Categorical(np.ones(3)/3)
        X = GaussianARD(0, 1, shape=(), plates=(3,4))
        Y = Gate(Z, X, gated_plate=-2)
        self.assertEqual(Y.plates, (4,))
        self.assertEqual(Y.dims, ( (), () ))

        # Fixed gating
        Z = 2
        X = GaussianARD(0, 1, shape=(2,), plates=(3,))
        Y = Gate(Z, X)
        self.assertEqual(Y.plates, ())
        self.assertEqual(Y.dims, ( (2,), (2,2) ))

        # Fixed X
        Z = Categorical(np.ones(3)/3)
        X = [1, 2, 3]
        Y = Gate(Z, X, moments=GaussianMoments(()))
        self.assertEqual(Y.plates, ())
        self.assertEqual(Y.dims, ( (), () ))

        # Do not accept non-negative cluster plates
        Z = Categorical(np.ones(3)/3)
        X = GaussianARD(0, 1, plates=(3,))
        self.assertRaises(ValueError,
                          Gate,
                          Z,
                          X,
                          gated_plate=0)

        # None of the parents have the cluster plate axis
        Z = Categorical(np.ones(3)/3)
        X = GaussianARD(0, 1)
        self.assertRaises(ValueError,
                          Gate,
                          Z,
                          X)

        # Inconsistent cluster plate
        Z = Categorical(np.ones(3)/3)
        X = GaussianARD(0, 1, plates=(2,))
        self.assertRaises(ValueError,
                          Gate,
                          Z,
                          X)

        pass


    def test_message_to_child(self):
        """
        Test the message to child of Gate node.
        """

        # Gating scalar node
        Z = 2
        X = GaussianARD([1,2,3], 1, shape=(), plates=(3,))
        Y = Gate(Z, X)
        u = Y._message_to_child()
        self.assertEqual(len(u), 2)
        self.assertAllClose(u[0], 3)
        self.assertAllClose(u[1], 3**2+1)

        # Fixed X
        Z = 2
        X = [1, 2, 3]
        Y = Gate(Z, X, moments=GaussianMoments(()))
        u = Y._message_to_child()
        self.assertEqual(len(u), 2)
        self.assertAllClose(u[0], 3)
        self.assertAllClose(u[1], 3**2)

        # Uncertain gating
        Z = Categorical([0.2,0.3,0.5])
        X = GaussianARD([1,2,3], 1, shape=(), plates=(3,))
        Y = Gate(Z, X)
        u = Y._message_to_child()
        self.assertAllClose(u[0], 0.2*1 + 0.3*2 + 0.5*3)
        self.assertAllClose(u[1], 0.2*2 + 0.3*5 + 0.5*10)

        # Plates in Z
        Z = [2, 0]
        X = GaussianARD([1,2,3], 1, shape=(), plates=(3,))
        Y = Gate(Z, X)
        u = Y._message_to_child()
        self.assertAllClose(u[0], [3, 1])
        self.assertAllClose(u[1], [10, 2])

        # Plates in X
        Z = 2
        X = GaussianARD([1,2,3], 1, shape=(), plates=(4,3,))
        Y = Gate(Z, X)
        u = Y._message_to_child()
        self.assertAllClose(np.ones(4)*u[0], np.ones(4)*3)
        self.assertAllClose(np.ones(4)*u[1], np.ones(4)*10)

        # Gating non-default plate
        Z = 2
        X = GaussianARD([[1],[2],[3]], 1, shape=(), plates=(3,4))
        Y = Gate(Z, X, gated_plate=-2)
        u = Y._message_to_child()
        self.assertAllClose(np.ones(4)*u[0], np.ones(4)*3)
        self.assertAllClose(np.ones(4)*u[1], np.ones(4)*10)

        # Gating non-scalar node
        Z = 2
        X = GaussianARD([1*np.ones(4),
                         2*np.ones(4),
                         3*np.ones(4)],
                        1,
                        shape=(4,), plates=(3,))
        Y = Gate(Z, X)
        u = Y._message_to_child()
        self.assertAllClose(u[0], 3*np.ones(4))
        self.assertAllClose(u[1], 9*np.ones((4,4)) + 1*np.identity(4))

        # Broadcasting the moments on the cluster axis
        Z = 2
        X = GaussianARD(1, 1, shape=(), plates=(3,))
        Y = Gate(Z, X)
        u = Y._message_to_child()
        self.assertEqual(len(u), 2)
        self.assertAllClose(u[0], 1)
        self.assertAllClose(u[1], 1**2+1)

        pass


    def test_message_to_parent(self):
        """
        Test the message to parents of Gate node.
        """

        # Unobserved and broadcasting
        Z = 2
        X = GaussianARD(0, 1, shape=(), plates=(3,))
        F = Gate(Z, X)
        Y = GaussianARD(F, 1)
        m = F._message_to_parent(0)
        self.assertEqual(len(m), 1)
        self.assertAllClose(m[0], 0*np.ones(3))
        m = F._message_to_parent(1)
        self.assertEqual(len(m), 2)
        self.assertAllClose(m[0]*np.ones(3), [0, 0, 0])
        self.assertAllClose(m[1]*np.ones(3), [0, 0, 0])

        # Gating scalar node
        Z = 2
        X = GaussianARD([1,2,3], 1, shape=(), plates=(3,))
        F = Gate(Z, X)
        Y = GaussianARD(F, 1)
        Y.observe(10)
        m = F._message_to_parent(0)
        self.assertAllClose(m[0], [10*1-0.5*2, 10*2-0.5*5, 10*3-0.5*10])
        m = F._message_to_parent(1)
        self.assertAllClose(m[0], [0, 0, 10])
        self.assertAllClose(m[1], [0, 0, -0.5])

        # Fixed X
        Z = 2
        X = [1,2,3]
        F = Gate(Z, X, moments=GaussianMoments(()))
        Y = GaussianARD(F, 1)
        Y.observe(10)
        m = F._message_to_parent(0)
        self.assertAllClose(m[0], [10*1-0.5*1, 10*2-0.5*4, 10*3-0.5*9])
        m = F._message_to_parent(1)
        self.assertAllClose(m[0], [0, 0, 10])
        self.assertAllClose(m[1], [0, 0, -0.5])

        # Uncertain gating
        Z = Categorical([0.2, 0.3, 0.5])
        X = GaussianARD([1,2,3], 1, shape=(), plates=(3,))
        F = Gate(Z, X)
        Y = GaussianARD(F, 1)
        Y.observe(10)
        m = F._message_to_parent(0)
        self.assertAllClose(m[0], [10*1-0.5*2, 10*2-0.5*5, 10*3-0.5*10])
        m = F._message_to_parent(1)
        self.assertAllClose(m[0], [0.2*10, 0.3*10, 0.5*10])
        self.assertAllClose(m[1], [-0.5*0.2, -0.5*0.3, -0.5*0.5])

        # Plates in Z
        Z = [2, 0]
        X = GaussianARD([1,2,3], 1, shape=(), plates=(3,))
        F = Gate(Z, X)
        Y = GaussianARD(F, 1)
        Y.observe([10, 20])
        m = F._message_to_parent(0)
        self.assertAllClose(m[0], [[10*1-0.5*2, 10*2-0.5*5, 10*3-0.5*10],
                                   [20*1-0.5*2, 20*2-0.5*5, 20*3-0.5*10]])
        m = F._message_to_parent(1)
        self.assertAllClose(m[0], [20, 0, 10])
        self.assertAllClose(m[1], [-0.5, 0, -0.5])

        # Plates in X
        Z = 2
        X = GaussianARD([[1,2,3], [4,5,6]], 1, shape=(), plates=(2,3,))
        F = Gate(Z, X)
        Y = GaussianARD(F, 1)
        Y.observe([10, 20])
        m = F._message_to_parent(0)
        self.assertAllClose(m[0], [10*1-0.5*2 + 20*4-0.5*17,
                                   10*2-0.5*5 + 20*5-0.5*26,
                                   10*3-0.5*10 + 20*6-0.5*37])
        m = F._message_to_parent(1)
        self.assertAllClose(m[0], [[0, 0, 10],
                                   [0, 0, 20]])
        self.assertAllClose(m[1]*np.ones((2,3)), [[0, 0, -0.5],
                                                  [0, 0, -0.5]])

        # Gating non-default plate
        Z = 2
        X = GaussianARD([[1],[2],[3]], 1, shape=(), plates=(3,1))
        F = Gate(Z, X, gated_plate=-2)
        Y = GaussianARD(F, 1)
        Y.observe([10])
        m = F._message_to_parent(0)
        self.assertAllClose(m[0], [10*1-0.5*2, 10*2-0.5*5, 10*3-0.5*10])
        m = F._message_to_parent(1)
        self.assertAllClose(m[0], [[0], [0], [10]])
        self.assertAllClose(m[1], [[0], [0], [-0.5]])

        # Gating non-scalar node
        Z = 2
        X = GaussianARD([[1,4],[2,5],[3,6]], 1, shape=(2,), plates=(3,))
        F = Gate(Z, X)
        Y = GaussianARD(F, 1)
        Y.observe([10,20])
        m = F._message_to_parent(0)
        self.assertAllClose(m[0], [10*1-0.5*2 + 20*4-0.5*17,
                                   10*2-0.5*5 + 20*5-0.5*26,
                                   10*3-0.5*10 + 20*6-0.5*37])
        m = F._message_to_parent(1)
        I = np.identity(2)
        self.assertAllClose(m[0], [[0,0], [0,0], [10,20]])
        self.assertAllClose(m[1], [0*I, 0*I, -0.5*I])

        # Broadcasting the moments on the cluster axis
        Z = 2
        X = GaussianARD(2, 1, shape=(), plates=(3,))
        F = Gate(Z, X)
        Y = GaussianARD(F, 1)
        Y.observe(10)
        m = F._message_to_parent(0)
        self.assertAllClose(m[0], [10*2-0.5*5, 10*2-0.5*5, 10*2-0.5*5])
        m = F._message_to_parent(1)
        self.assertAllClose(m[0], [0, 0, 10])
        self.assertAllClose(m[1], [0, 0, -0.5])

        pass


    def test_mask_to_parent(self):
        """
        Test the mask handling in Gate node
        """

        X = GaussianARD(2, 1, shape=(4, 5), plates=(3, 2))
        F = Gate([0, 0, 1], X)

        self.assertAllClose(
            F._compute_weights_to_parent(0, [True, False, False]),
            [True, False, False]
        )
        self.assertAllClose(
            F._compute_weights_to_parent(1, [True, False, False]),
            [[True], [False], [False]]
        )

        pass



################################################################################
# Copyright (C) 2014 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


"""
"""

import numpy as np

from bayespy.utils import misc

from .node import Node, Moments
from .deterministic import Deterministic
from .categorical import CategoricalMoments
from .concatenate import Concatenate


class Gate(Deterministic):
    """
    Deterministic gating of one node.

    Gating is performed over one plate axis.

    Note: You should not use gating for several variables which parents of a
    same node if the gates use the same gate assignments.  In such case, the
    results will be wrong.  The reason is a general one: A stochastic node may
    not be a parent of another node via several paths unless at most one path
    has no other stochastic nodes between them.
    """

    def __init__(self, Z, X, gated_plate=-1, moments=None, **kwargs):
        """
        Constructor for the gating node.

        Parameters
        ----------

        Z : Categorical-like node
           A variable which chooses the index along the gated plate axis

        X : node
           The node whose plate axis is gated

        gated_plate : int (optional)
           The index of the plate axis to be gated (by default, -1, that is,
           the last axis).
        """

        if gated_plate >= 0:
            raise ValueError("Cluster plate must be negative integer")
        self.gated_plate = gated_plate

        if moments is not None:
            X = self._ensure_moments(
                X,
                moments.__class__,
                **moments.get_instance_conversion_kwargs()
            )

        if not isinstance(X, Node):
            raise ValueError("X must be a node or moments should be provided")

        X_moments = X._moments
        self._moments = X_moments
        dims = X.dims
        if len(X.plates) < abs(gated_plate):
            raise ValueError("The gated node does not have a plate axis is "
                             "gated")
        K = X.plates[gated_plate]

        Z = self._ensure_moments(Z, CategoricalMoments, categories=K)

        self._parent_moments = (Z._moments, X_moments)

        if Z.dims != ( (K,), ):
            raise ValueError("Inconsistent number of clusters")

        self.K = K

        super().__init__(Z, X, dims=dims, **kwargs)


    def _compute_moments(self, u_Z, u_X):
        """
        """

        u = []
        for i in range(len(u_X)):
            # Make the moments of Z and X broadcastable and move the gated plate
            # to be the last axis in the moments, then sum-product over that
            # axis
            ndim = len(self.dims[i])
            z = misc.add_trailing_axes(u_Z[0], ndim)
            z = misc.moveaxis(z, -ndim-1, -1)
            gated_axis = self.gated_plate - ndim
            if np.ndim(u_X[i]) < abs(gated_axis):
                x = misc.add_trailing_axes(u_X[i], 1)
            else:
                x = misc.moveaxis(u_X[i], gated_axis, -1)
            ui = misc.sum_product(z, x, axes_to_sum=-1)
            u.append(ui)
        return u


    def _compute_message_to_parent(self, index, m_child, u_Z, u_X):
        """
        """
        if index == 0:
            m0 = 0
            # Compute Child * X, sum over variable axes and move the gated axis
            # to be the last.  Need to do some shape changing in order to make
            # Child and X to broadcast properly.
            for i in range(len(m_child)):
                ndim = len(self.dims[i])
                c = m_child[i][...,None]
                c = misc.moveaxis(c, -1, -ndim-1)
                gated_axis = self.gated_plate - ndim
                x = u_X[i]
                if np.ndim(x) < abs(gated_axis):
                    x = np.expand_dims(x, -ndim-1)
                else:
                    x = misc.moveaxis(x, gated_axis, -ndim-1)
                axes = tuple(range(-ndim, 0))
                m0 = m0 + misc.sum_product(c, x, axes_to_sum=axes)

            # Make sure the variable axis does not use broadcasting
            m0 = m0 * np.ones(self.K)

            # Send the message
            m = [m0]
            return m

        elif index == 1:

            m = []
            for i in range(len(m_child)):
                # Make the moments of Z and the message from children
                # broadcastable. The gated plate is handled as the last axis in
                # the arrays and moved to the correct position at the end.

                # Add variable axes to Z moments
                ndim = len(self.dims[i])
                z = misc.add_trailing_axes(u_Z[0], ndim)
                z = misc.moveaxis(z, -ndim-1, -1)
                # Axis index of the gated plate
                gated_axis = self.gated_plate - ndim
                # Add the gate axis to the message from the children
                c = misc.add_trailing_axes(m_child[i], 1)
                # Compute the message to parent
                mi = z * c
                # Add extra axes if necessary
                if np.ndim(mi) < abs(gated_axis):
                    mi = misc.add_leading_axes(mi,
                                                abs(gated_axis) - np.ndim(mi))
                # Move the axis to the correct position
                mi = misc.moveaxis(mi, -1, gated_axis)
                m.append(mi)

            return m

        else:
            raise ValueError("Invalid parent index")


    def _compute_weights_to_parent(self, index, weights):
        """
        """
        if index == 0:
            return weights
        elif index == 1:
            if self.gated_plate >= 0:
                raise ValueError("Gated plate axis must be negative")
            return (
                np.expand_dims(weights, axis=self.gated_plate)
                if np.ndim(weights) >= abs(self.gated_plate) else
                weights
            )
        else:
            raise ValueError("Invalid parent index")


    def _compute_plates_to_parent(self, index, plates):
        """
        """
        if index == 0:
            return plates
        elif index == 1:
            plates = list(plates)
            # Add the cluster plate axis
            if self.gated_plate < 0:
                knd = len(plates) + self.gated_plate + 1
            else:
                raise RuntimeError("Cluster plate axis must be negative")
            plates.insert(knd, self.K)

            return tuple(plates)
        else:
            raise ValueError("Invalid parent index")


    def _compute_plates_from_parent(self, index, plates):
        """
        """
        if index == 0:
            return plates
        elif index == 1:
            plates = list(plates)
            # Remove the cluster plate, if the parent has it
            if len(plates) >= abs(self.gated_plate):
                plates.pop(self.gated_plate)
            return tuple(plates)
        else:
            raise ValueError("Invalid parent index")


def Choose(z, *nodes):
    """Choose plate elements from nodes based on a categorical variable.

    For instance:

    .. testsetup::

       from bayespy.nodes import *

    .. code-block:: python

       >>> import bayespy as bp
       >>> z = [0, 0, 2, 1]
       >>> x0 = bp.nodes.GaussianARD(0, 1)
       >>> x1 = bp.nodes.GaussianARD(10, 1)
       >>> x2 = bp.nodes.GaussianARD(20, 1)
       >>> x = bp.nodes.Choose(z, x0, x1, x2)
       >>> print(x.get_moments()[0])
       [ 0.  0. 20. 10.]

    This is basically just a thin wrapper over applying Gate node over the
    concatenation of the nodes.
    """
    categories = len(nodes)
    z = Deterministic._ensure_moments(
        z,
        CategoricalMoments,
        categories=categories
    )
    nodes = [node[...,None] for node in nodes]
    combined = Concatenate(*nodes)
    return Gate(z, combined)



################################################################################
# Copyright (C) 2013 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


r"""
General functions random sampling and distributions.
"""

import numpy as np
from scipy import special

from . import linalg
from . import misc


def intervals(N, length, amount=1, gap=0):
    r"""
    Return random non-overlapping parts of a sequence.

    For instance, N=16, length=2 and amount=4:
    [0, |1, 2|, 3, 4, 5, |6, 7|, 8, 9, |10, 11|, |12, 13|, 14, 15]
    that is,
    [1,2,6,7,10,11,12,13]

    However, the function returns only the indices of the beginning of the
    sequences, that is, in the example:
    [1,6,10,12]
    """

    if length * amount + gap * (amount-1) > N:
        raise ValueError("Too short sequence")

    # In practice, we draw the sizes of the gaps between the sequences
    total_gap = N - length*amount - gap*(amount-1)
    gaps = np.random.multinomial(total_gap, np.ones(amount+1)/(amount+1))

    # And then we get the beginning index of each sequence
    intervals = np.cumsum(gaps[:-1]) + np.arange(amount)*(length+gap)

    return intervals

def mask(*shape, p=0.5):
    r"""
    Return a boolean array of the given shape.

    Parameters
    ----------
    d0, d1, ..., dn : int
        Shape of the output.
    p : value in range [0,1]
        A probability that the elements are `True`.
    """
    return np.random.rand(*shape) < p

def wishart(nu, V):
    r"""
    Draw a random sample from the Wishart distribution.

    Parameters
    ----------
    nu : int
    """
    # TODO/FIXME: Are these correct..
    D = np.shape(V)[0]
    if nu < D:
        raise ValueError("Degrees of freedom must be equal or greater than the "
                         "dimensionality of the matrix.")
    X = np.random.multivariate_normal(np.zeros(D), V, size=nu)
    return np.dot(X, X.T)

wishart_rand = wishart

def invwishart_rand(nu, V):
    # TODO/FIXME: Are these correct..
    return np.linalg.inv(wishart_rand(nu, V))

def covariance(D, size=(), nu=None):
    r"""
    Draw a random covariance matrix.

    Draws from inverse-Wishart distribution. The distribution of each element is
    independent of the dimensionality of the matrix.

    C ~ Inv-W(I, D)

    Parameters
    ----------
    D : int
        Dimensionality of the covariance matrix.

    Returns:
    --------
    C : (D,D) ndarray
        Positive-definite symmetric :math:`D\times D` matrix.
    """

    if nu is None:
        nu = D

    if nu < D:
        raise ValueError("nu must be greater than or equal to D")

    try:
        size = tuple(size)
    except TypeError:
        size = (size,)
    shape = size + (D,nu)
    C = np.random.randn(*shape)
    C = linalg.dot(C, np.swapaxes(C, -1, -2)) / nu
    return linalg.inv(C)
#return np.linalg.inv(np.dot(C, C.T))

def correlation(D):
    r"""
    Draw a random correlation matrix.
    """
    X = np.random.randn(D,D);
    s = np.sqrt(np.sum(X**2, axis=-1, keepdims=True))
    X = X / s
    return np.dot(X, X.T)


def gaussian_logpdf(yVy, yVmu, muVmu, logdet_V, D):
    r"""
    Log-density of a Gaussian distribution.

    :math:`\mathcal{G}(\mathbf{y}|\boldsymbol{\mu},\mathbf{V}^{-1})`

    Parameters
    -----------
    yVy : ndarray or double
        :math:`\mathbf{y}^T\mathbf{Vy}`
    yVmu : ndarray or double
        :math:`\mathbf{y}^T\mathbf{V}\boldsymbol{\mu}`
    muVmu : ndarray or double
        :math:`\boldsymbol{\mu}^T\mathbf{V}\boldsymbol{\mu}`
    logdet_V : ndarray or double
        Log-determinant of the precision matrix, :math:`\log|\mathbf{V}|`.
    D : int
        Dimensionality of the distribution.
    """
    return -0.5*yVy + yVmu - 0.5*muVmu + 0.5*logdet_V - 0.5*D*np.log(2*np.pi)

def gaussian_entropy(logdet_V, D):
    r"""
    Compute the entropy of a Gaussian distribution.

    If you want to get the gradient, just let each parameter be a gradient of
    that term.

    Parameters
    ----------
    logdet_V : ndarray or double
        The log-determinant of the precision matrix.
    D : int
        The dimensionality of the distribution.
    """
    return -0.5*logdet_V + 0.5*D + 0.5*D*np.log(2*np.pi)

def gamma_logpdf(bx, logx, a_logx, a_logb, gammaln_a):
    r"""
    Log-density of :math:`\mathcal{G}(x|a,b)`.
    
    If you want to get the gradient, just let each parameter be a gradient of
    that term.

    Parameters
    ----------
    bx : ndarray
        :math:`bx`
    logx : ndarray
        :math:`\log(x)`
    a_logx : ndarray
        :math:`a \log(x)`
    a_logb : ndarray
        :math:`a \log(b)`
    gammaln_a : ndarray
        :math:`\log\Gamma(a)`
    """
    return a_logb - gammaln_a + a_logx - logx - bx
#def gamma_logpdf(a, log_b, gammaln_a, 

def gamma_entropy(a, log_b, gammaln_a, psi_a, a_psi_a):
    r"""
    Entropy of :math:`\mathcal{G}(a,b)`.

    If you want to get the gradient, just let each parameter be a gradient of
    that term.

    Parameters
    ----------
    a : ndarray
        :math:`a`
    log_b : ndarray
        :math:`\log(b)`
    gammaln_a : ndarray
        :math:`\log\Gamma(a)`
    psi_a : ndarray
        :math:`\psi(a)`
    a_psi_a : ndarray
        :math:`a\psi(a)`
    """
    return a - log_b + gammaln_a + psi_a - a_psi_a

def orth(D):
    r"""
    Draw random orthogonal matrix.
    """
    Q = np.random.randn(D,D)
    (Q, _) = np.linalg.qr(Q)
    return Q

def svd(s):
    r"""
    Draw a random matrix given its singular values.
    """
    D = len(s)
    U = orth(D) * s
    V = orth(D)
    return np.dot(U, V.T)
    
def sphere(N=1):
    r"""
    Draw random points uniformly on a unit sphere.

    Returns (latitude,longitude) in degrees.
    """
    lon = np.random.uniform(-180, 180, N)
    lat = (np.arccos(np.random.uniform(-1, 1, N)) * 180 / np.pi) - 90
    return (lat, lon)


def bernoulli(p, size=None):
    r"""
    Draw random samples from the Bernoulli distribution.
    """
    if isinstance(size, int):
        size = (size,)
    if size is None:
        size = np.shape(p)
    return (np.random.rand(*size) < p)


def categorical(p, size=None):
    r"""
    Draw random samples from a categorical distribution.
    """
    if size is None:
        size = np.shape(p)[:-1]
    if isinstance(size, int):
        size = (size,)

    if np.any(np.asanyarray(p)<0):
        raise ValueError("Array contains negative probabilities")

    if not misc.is_shape_subset(np.shape(p)[:-1], size):
        raise ValueError("Probability array shape and requested size are "
                         "inconsistent")

    size = tuple(size)

    # Normalize probabilities
    p = p / np.sum(p, axis=-1, keepdims=True)

    # Compute cumulative probabilities (p_1, p_1+p_2, ..., p_1+...+p_N):
    P = np.cumsum(p, axis=-1)

    # Draw samples from interval [0,1]
    x = np.random.rand(*size)

    # For simplicity, repeat p to the size of the output (plus probability axis)
    K = np.shape(p)[-1]
    P = P * np.ones(tuple(size)+(K,))

    if size == ():
        z = np.searchsorted(P, x)
    else:
        # Seach the indices
        z = np.zeros(size)
        inds = misc.nested_iterator(size)
        for ind in inds:
            z[ind] = np.searchsorted(P[ind], x[ind])

    return z.astype(int)


def multinomial(n, p, size=None):

    plates_n = np.shape(n)
    plates_p = np.shape(p)[:-1]
    k = np.shape(p)[-1]

    if size is None:
        size = misc.broadcasted_shape(plates_n, plates_p)

    if not misc.is_shape_subset(plates_n, size):
        raise ValueError("Shape of n does not broadcast to the given size")

    if not misc.is_shape_subset(plates_p, size):
        raise ValueError("Shape of p does not broadcast to the given size")

    # This isn't a very efficient implementation. One could use NumPy's
    # multinomial once for all those plates for which n and p is the same.

    n = np.broadcast_to(n, size)
    p = np.broadcast_to(p, size + (k,))

    x = np.empty(size + (k,))

    for i in misc.nested_iterator(size):
        x[i] = np.random.multinomial(n[i], p[i])

    return x.astype(int)


def gamma(a, b, size=None):
    x = np.random.gamma(a, b, size=size)
    if np.any(x == 0):
        raise RuntimeError(
            "Numerically zero samples. Try using a larger shape parameter in "
            "the gamma distribution."
        )
    return x


def dirichlet(alpha, size=None):
    r"""
    Draw random samples from the Dirichlet distribution.
    """
    if isinstance(size, int):
        size = (size,)
    if size is None:
        size = np.shape(alpha)
    else:
        size = size + np.shape(alpha)[-1:]
    p = np.random.gamma(alpha, size=size)
    sump = np.sum(p, axis=-1, keepdims=True)
    if np.any(sump == 0):
        raise RuntimeError(
            "Numerically zero samples. Try using a larger Dirichlet "
            "concentration parameter value."
        )
    p /= sump
    return p


def logodds_to_probability(x):
    r"""
    Solves p from log(p/(1-p))
    """
    return 1 / (1 + np.exp(-x)) 
    

def alpha_beta_recursion(logp0, logP):
    r"""
    Compute alpha-beta recursion for Markov chain

    Initial state log-probabilities are in `p0` and state transition
    log-probabilities are in P. The probabilities do not need to be scaled to
    sum to one, but they are interpreted as below:

    logp0 = log P(z_0) + log P(y_0|z_0)
    logP[...,n,:,:] = log P(z_{n+1}|z_n) + log P(y_{n+1}|z_{n+1})
    """

    logp0 = misc.atleast_nd(logp0, 1)
    logP = misc.atleast_nd(logP, 3)
    
    D = np.shape(logp0)[-1]
    N = np.shape(logP)[-3]
    plates = misc.broadcasted_shape(np.shape(logp0)[:-1], np.shape(logP)[:-3])

    if np.shape(logP)[-2:] != (D,D):
        raise ValueError("Dimension mismatch %s != %s"
                         % (np.shape(logP)[-2:],
                            (D,D)))

    #
    # Run the recursion algorithm
    #

    # Allocate memory
    logalpha = np.zeros(plates+(N,D))
    logbeta = np.zeros(plates+(N,D))
    g = np.zeros(plates)

    # Forward recursion
    logalpha[...,0,:] = logp0
    for n in range(1,N):
        # Compute: P(z_{n-1},z_n|x_1,...,x_n)
        v = logalpha[...,n-1,:,None] + logP[...,n-1,:,:]
        c = misc.logsumexp(v, axis=(-1,-2))
        # Sum over z_{n-1} to get: log P(z_n|x_1,...,x_n)
        logalpha[...,n,:] = misc.logsumexp(v - c[...,None,None], axis=-2)
        g -= c

    # Compute the normalization of the last term
    v = logalpha[...,N-1,:,None] + logP[...,N-1,:,:]
    g -= misc.logsumexp(v, axis=(-1,-2))

    # Backward recursion 
    logbeta[...,N-1,:] = 0
    for n in reversed(range(N-1)):
        v = logbeta[...,n+1,None,:] + logP[...,n+1,:,:]
        c = misc.logsumexp(v, axis=(-1,-2))
        logbeta[...,n,:] = misc.logsumexp(v - c[...,None,None], axis=-1)

    v = logalpha[...,:,:,None] + logbeta[...,:,None,:] + logP[...,:,:,:]
    c = misc.logsumexp(v, axis=(-1,-2))
    zz = np.exp(v - c[...,None,None])

    # The logsumexp normalization is not numerically accurate, so do
    # normalization again:
    zz /= np.sum(zz, axis=(-1,-2), keepdims=True)

    z0 = np.sum(zz[...,0,:,:], axis=-1)
    z0 /= np.sum(z0, axis=-1, keepdims=True)

    return (z0, zz, g)


def gaussian_gamma_to_t(mu, Cov, a, b, ndim=1):
    r"""
    Integrates gamma distribution to obtain parameters of t distribution
    """
    alpha = a/b
    nu = 2*a
    S = Cov / misc.add_trailing_axes(alpha, 2*ndim)
    return (mu, S, nu)


def t_logpdf(z2, logdet_cov, nu, D):
    r"""
    """
    return (special.gammaln((nu+D)/2)
            - special.gammaln(nu/2)
            - 0.5 * D * np.log(nu*np.pi)
            - 0.5 * logdet_cov
            - 0.5 * (nu+D) * np.log(1 + z2/nu))



################################################################################
# Copyright (C) 2011-2016 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################

from . import utils
from . import inference
from . import nodes

try:
    from . import plot
except ImportError:
    # Matplotlib not available
    pass

from ._meta import __author__, __copyright__, __contact__, __license__

from . import _version
__version__ = _version.get_versions()['version']



################################################################################
# Copyright (C) 2011-2014 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


"""
General numerical functions and methods.

"""

import itertools
import numpy as np
import scipy as sp
#import scipy.linalg.decomp_cholesky as decomp
import scipy.linalg as linalg
import scipy.special as special
import scipy.optimize as optimize
import scipy.sparse as sparse
#import scikits.sparse.cholmod as cholmod

# THIS IS SOME NEW GENERALIZED UFUNC FOR LINALG FEATURE, NOT IN OFFICIAL NUMPY
# REPO YET
#import numpy.linalg._gufuncs_linalg as gula
#import numpy.core.gufuncs_linalg as gula

from . import misc


def chol(C, ndim=1):
    if sparse.issparse(C):
        if ndim != 1:
            raise NotImplementedError()
        # Sparse Cholesky decomposition (returns a Factor object)
        return cholmod.cholesky(C)
    else:
        # Computes Cholesky decomposition for a collection of matrices.
        # The last 2*ndim axes of C are considered as the matrix.
        if ndim == 0:
            return np.sqrt(C)
        shape_original = np.shape(C)[-ndim:]
        C = (
            C if ndim == 1 else
            misc.flatten_axes(C, ndim, ndim)
        )
        if np.shape(C)[-1] != np.shape(C)[-2]:
            raise ValueError("Not square matrix w.r.t. ndim sense")
        U = np.empty(np.shape(C))
        for i in misc.nested_iterator(np.shape(U)[:-2]):
            try:
                # Handle (0, 0) matrices. See:
                # https://github.com/scipy/scipy/issues/8056
                U[i] = (
                    np.empty((0,0)) if np.size(C[i]) == 0 else
                    linalg.cho_factor(C[i])[0]
                )
            except np.linalg.linalg.LinAlgError:
                raise Exception("Matrix not positive definite")
        return (
            U if ndim == 1 else
            misc.reshape_axes(U, shape_original, shape_original)
        )


def chol_solve(U, b, out=None, matrix=False, ndim=1):
    if isinstance(U, np.ndarray) or np.isscalar(U):
        if sparse.issparse(b):
            b = b.toarray()

        if ndim == 0:
            return (b / U) / U

        shape = np.shape(U)[-ndim:]
        U = (
            U if ndim == 1 else
            misc.flatten_axes(U, ndim, ndim)
        )

        if matrix:
            shape_b = np.shape(b)[-ndim:]
            B = (
                b if ndim == 1 else
                misc.flatten_axes(b, ndim, ndim)
            )
            B = transpose(B, ndim=1)
            U = U[...,None,:,:]
        else:
            B = (
                b if ndim == 1 else
                misc.flatten_axes(b, ndim)
            )

        # Allocate memory
        sh_u = U.shape[:-2]
        sh_b = B.shape[:-1]
        l_u = len(sh_u)
        l_b = len(sh_b)

        # Check which axis are iterated over with B along with U
        ind_b = [slice(None)] * l_b
        l_min = min(l_u, l_b)
        jnd_b = tuple(i for i in range(-l_min,0) if sh_b[i]==sh_u[i])

        if out == None:
            # Shape of the result (broadcasting rules)
            sh = misc.broadcasted_shape(sh_u, sh_b)
            #out = np.zeros(np.shape(B))
            out = np.zeros(sh + B.shape[-1:])

        for i in misc.nested_iterator(np.shape(U)[:-2]):

            # The goal is to run Cholesky solver once for all vectors of B
            # for which the matrices of U are the same (according to the
            # broadcasting rules). Thus, we collect all the axes of B for
            # which U is singleton and form them as a 2-D matrix and then
            # run the solver once.

            # Select those axes of B for which U and B are not singleton
            for j in jnd_b:
                ind_b[j] = i[j]

            # Collect all the axes for which U is singleton
            b = B[tuple(ind_b) + (slice(None),)]

            # Reshape it to a 2-D (or 1-D) array
            orig_shape = b.shape
            if b.ndim > 1:
                b = b.reshape((-1, b.shape[-1]))

            # slice(None) to all preceeding axes and ellipsis for the last
            # axis:
            if len(ind_b) < len(sh):
                ind_out = (slice(None),) + tuple(ind_b) + (slice(None),)
            else:
                ind_out = tuple(ind_b) + (slice(None),)

            out[ind_out] = (
                # Handle (0, 0) matrices. See:
                # https://github.com/scipy/scipy/issues/8056
                np.empty(orig_shape) if np.size(U[i]) == 0 else
                linalg.cho_solve(
                    (U[i], False),
                    b.T
                ).T.reshape(orig_shape)
            )

        if matrix:
            out = transpose(out, ndim=1)
            out = (
                out if ndim == 1 else
                misc.reshape_axes(out, shape, shape_b)
            )
        else:
            out = (
                out if ndim == 1 else
                misc.reshape_axes(out, shape)
            )

        return out

    elif isinstance(U, cholmod.Factor):
        if ndim != 1:
            raise NotImplementedError()
        if matrix:
            raise NotImplementedError()
        if sparse.issparse(b):
            b = b.toarray()
        return U.solve_A(b)
    else:
        raise ValueError("Unknown type of Cholesky factor")


def chol_inv(U, ndim=1):
    if isinstance(U, np.ndarray) or np.isscalar(U):
        if ndim == 0:
            return (1 / U) / U
        shape = np.shape(U)[-ndim:]
        U = (
            U if ndim == 1 else
            misc.flatten_axes(U, ndim, ndim)
        )
        # Allocate memory
        V = np.tile(np.identity(np.shape(U)[-1]), np.shape(U)[:-2]+(1,1))
        for i in misc.nested_iterator(np.shape(U)[:-2]):
            V[i] = (
                # Handle (0, 0) matrices. See:
                # https://github.com/scipy/scipy/issues/8056
                np.empty((0, 0)) if np.size(V[i]) == 0 else
                linalg.cho_solve(
                    (U[i], False),
                    V[i],
                    overwrite_b=True # This would need Fortran order
                )
            )
        V = (
            V if ndim == 1 else
            misc.reshape_axes(V, shape, shape)
        )
        return V

    elif isinstance(U, cholmod.Factor):
        raise NotImplementedError
        if ndim != 1:
            raise NotImplementedError()
    else:
        raise ValueError("Unknown type of Cholesky factor")

def chol_logdet(U, ndim=1):
    if isinstance(U, np.ndarray) or np.isscalar(U):
        if ndim == 0:
            return 2 * np.log(U)
        U = (
            U if ndim == 1 else
            misc.flatten_axes(U, ndim, ndim)
        )
        return 2*np.sum(np.log(np.einsum('...ii->...i',U)), axis=-1)
    elif isinstance(U, cholmod.Factor):
        if ndim != 1:
            raise NotImplementedError()
        return np.sum(np.log(U.D()))
    else:
        raise ValueError("Unknown type of Cholesky factor")

def logdet_chol(U):
    if isinstance(U, np.ndarray):
        # Computes Cholesky decomposition for a collection of matrices.
        return 2*np.sum(np.log(np.einsum('...ii->...i', U)), axis=(-1,))
    elif isinstance(U, cholmod.Factor):
        return np.sum(np.log(U.D()))


def logdet_tri(R):
    """
    Logarithm of the absolute value of the determinant of a triangular matrix.
    """
    return np.sum(np.log(np.abs(np.einsum('...ii->...i', R))))


def logdet_cov(C, ndim=1):
    return chol_logdet(chol(C, ndim=ndim), ndim=ndim)


def solve_triangular(U, B, ndim=1, **kwargs):
    if ndim != 1:
        raise NotImplementedError("Not yet implemented for ndim!=1")
    # Allocate memory
    U = np.atleast_2d(U)
    B = np.atleast_1d(B)
    sh_u = U.shape[:-2]
    sh_b = B.shape[:-1]
    l_u = len(sh_u)
    l_b = len(sh_b)

    # Check which axis are iterated over with B along with U
    ind_b = [slice(None)] * l_b
    l_min = min(l_u, l_b)
    jnd_b = tuple(i for i in range(-l_min,0) if sh_b[i]==sh_u[i])

    # Shape of the result (broadcasting rules)
    sh = misc.broadcasted_shape(sh_u, sh_b)
    out = np.zeros(sh + B.shape[-1:])
    for i in misc.nested_iterator(np.shape(U)[:-2]):

        # The goal is to run triangular solver once for all vectors of
        # B for which the matrices of U are the same (according to the
        # broadcasting rules). Thus, we collect all the axes of B for
        # which U is singleton and form them as a 2-D matrix and then
        # run the solver once.

        # Select those axes of B for which U and B are not singleton
        for j in jnd_b:
            ind_b[j] = i[j]

        # Collect all the axes for which U is singleton
        b = B[tuple(ind_b) + (slice(None),)]

        # Reshape it to a 2-D (or 1-D) array
        orig_shape = b.shape
        if b.ndim > 1:
            b = b.reshape((-1, b.shape[-1]))

        # slice(None) to all preceeding axes and ellipsis for the last
        # axis:
        if len(ind_b) < len(sh):
            ind_out = (slice(None),) + tuple(ind_b) + (slice(None),)
        else:
            ind_out = tuple(ind_b) + (slice(None),)

        out[ind_out] = linalg.solve_triangular(U[i],
                                               b.T,
                                               **kwargs).T.reshape(orig_shape)

    return out




def inner(*args, ndim=1):
    """
    Compute inner product.

    The number of arrays is arbitrary.  The number of dimensions is arbitrary.
    """
    axes = tuple(range(-ndim,0))
    return misc.sum_product(*args, axes_to_sum=axes)


def outer(A, B, ndim=1):
    """
    Computes outer product over the last axes of A and B.

    The other axes are broadcasted. Thus, if A has shape (..., N) and B has
    shape (..., M), then the result has shape (..., N, M).

    Using the argument `ndim` it is possible to change that how many axes
    trailing axes are used for the outer product. For instance, if ndim=3, A and
    B have shapes (...,N1,N2,N3) and (...,M1,M2,M3), the result has shape
    (...,N1,M1,N2,M2,N3,M3).
    """
    if not isinstance(ndim, int) or ndim < 0:
        raise ValueError('ndim must be non-negative integer')
    if ndim > 0:
        if ndim > np.ndim(A):
            raise ValueError('Argument ndim larger than ndim of the first '
                             'parameter')
        if ndim > np.ndim(B):
            raise ValueError('Argument ndim larger than ndim of the second '
                             'parameter')
        shape_A = np.shape(A) + (1,)*ndim
        shape_B = np.shape(B)[:-ndim] + (1,)*ndim + np.shape(B)[-ndim:]
        A = np.reshape(A, shape_A)
        B = np.reshape(B, shape_B)
    return np.asanyarray(A) * np.asanyarray(B)


def _dot(A, B):
    """
    Dot product which handles broadcasting properly.

    Future NumPy will have a better built-in implementation for this.
    """
    A_plates = np.shape(A)[:-2]
    B_plates = np.shape(B)[:-2]
    M = np.shape(A)[-2]
    N = np.shape(B)[-1]
    Y_plates = misc.broadcasted_shape(A_plates, B_plates)
    if Y_plates == ():
        return np.dot(A, B)
    indices = misc.nested_iterator(Y_plates)
    Y_shape = Y_plates + (M, N)
    Y = np.zeros(Y_shape)
    for i in indices:
        Y[i] = np.dot(A[misc.safe_indices(i, A_plates)],
                      B[misc.safe_indices(i, B_plates)])
    return Y

def dot(*arrays):
    """
    Compute matrix-matrix product.

    You can give multiple arrays, the dot product is computed from left to
    right: A1*A2*A3*...*AN. The dot product is computed over the last two axes
    of each arrays. All other axes must be broadcastable.
    """
    if len(arrays) == 0:
        return 0
    else:
        Y = np.asanyarray(arrays[0])
        for X in arrays[1:]:
            X = np.asanyarray(X)
            if np.ndim(Y) < 2 or np.ndim(X) < 2:
                raise ValueError("Must be at least 2-D arrays")
            if np.shape(Y)[-1] != np.shape(X)[-2]:
                raise ValueError("Dimensions do not match")
            # Replace this with numpy.dot when NumPy implements broadcasting in dot
            Y = _dot(Y, X)
            #Y = np.einsum('...ik,...kj->...ij', Y, X)
            #Y = gula.matrix_multiply(Y, X)
        return Y

def tracedot(A, B):
    """
    Computes trace(A*B).
    """
    return np.einsum('...ij,...ji->...', A, B)


def inv(A, ndim=1):
    """
    General array inversion.

    Supports broadcasting and inversion of multidimensional arrays.  For
    instance, an array with shape (4,3,2,3,2) could mean that there are four
    (3*2) x (3*2) matrices to be inverted. This can be done by inv(A, ndim=2).
    For inverting scalars, ndim=0. For inverting matrices, ndim=1.
    """
    A = np.asanyarray(A)
    if ndim == 0:
        return 1 / A
    elif ndim == 1:
        return np.linalg.inv(A)
    else:
        raise NotImplementedError()


def mvdot(A, b, ndim=1):
    """
    Compute matrix-vector product.

    Applies broadcasting.
    """
    # TODO/FIXME: A bug in inner1d:
    # https://github.com/numpy/numpy/issues/3338
    #
    # b = np.asanyarray(b)
    # return gula.inner1d(A, b[...,np.newaxis,:])
    #
    # Use einsum instead:
    if ndim > 0:
        b = misc.add_axes(b, num=ndim, axis=-1-ndim)

    return inner(A, b, ndim=ndim)
    ## if ndim != 1:
    ##     raise NotImplementedError("mvdot not yet implemented for ndim!=1")

    ## return _dot(A, b[...,None])[...,0]
    ## #return np.einsum('...ik,...k->...i', A, b)

def mmdot(A, B, ndim=1):
    """
    Compute matrix-matrix product.

    Applies broadcasting.
    """
    if ndim == 0:
        return A * B
    elif ndim == 1:
        return _dot(A, B)
    else:
        raise Exception("mmdot not yet implemented for ndim>1")
    #return np.einsum('...ik,...kj->...ij', A, B)

def transpose(X, ndim=1):
    """
    Transpose the matrix.
    """
    for n in range(ndim):
        X = np.swapaxes(X, -1-n, -1-ndim-n)
    return X
    ## if ndim != 1:
    ##     raise Exception("transpose not yet implemented for ndim!=1")
    ## return np.swapaxes(X, -1, -2)

def m_dot(A,b):
    raise DeprecationWarning()
    # Compute matrix-vector product over the last two axes of A and
    # the last axes of b.  Other axes are broadcasted. If A has shape
    # (..., M, N) and b has shape (..., N), then the result has shape
    # (..., M)

    #b = reshape(b, shape(b)[:-1] + (1,) + shape(b)[-1:])
    #return np.dot(A, b)
    return np.einsum('...ik,...k->...i', A, b)
    # TODO: Use einsum!!
    #return np.sum(A*b[...,np.newaxis,:], axis=(-1,))

def block_banded_solve(A, B, y):
    """
    Invert symmetric, banded, positive-definite matrix.

    A contains the diagonal blocks.

    B contains the superdiagonal blocks (their transposes are the
    subdiagonal blocks).

    Shapes:
    A: (...,   N, D, D)
    B: (..., N-1, D, D)
    y: (...,   N,    D)

    The algorithm is basically LU decomposition.

    Computes only the diagonal and super-diagonal blocks of the
    inverse. The true inverse is dense, in general.

    Assume each block has the same size.

    Return:
    * inverse blocks
    * solution to the system
    * log-determinant
    """

    # Number of time instance and dimensionality
    N = np.shape(y)[-2]
    D = np.shape(y)[-1]

    # Check the shape of the diagonal blocks
    if np.shape(A)[-3] != N:
        raise ValueError("The number of diagonal blocks is incorrect")
    if np.shape(A)[-2:] != (D,D):
        raise ValueError("The diagonal blocks have wrong shape")

    # Check the shape of the super-diagonal blocks
    if np.shape(B)[-3] != N-1:
        raise ValueError("The number of super-diagonal blocks is incorrect")
    if np.shape(B)[-2:] != (D,D):
        raise ValueError("The diagonal blocks have wrong shape")

    plates_VC = misc.broadcasted_shape(np.shape(A)[:-3],
                                       np.shape(B)[:-3])
    plates_y = misc.broadcasted_shape(plates_VC,
                                      np.shape(y)[:-2])

    V = np.empty(plates_VC+(N,D,D))
    C = np.empty(plates_VC+(N-1,D,D))
    x = np.empty(plates_y+(N,D))

    #
    # Forward recursion
    #

    # In the forward recursion, store the Cholesky factor in V. So you
    # don't need to recompute them in the backward recursion.

    # TODO: This whole algorithm could be implemented as in-place operation.
    # Might be a nice feature (optional?)

    # TODO/FIXME: chol_solve has quite a high overhead because it uses shape
    # manipulations. Use some more raw method instead.

    x[...,0,:] = y[...,0,:]
    V[...,0,:,:] = chol(A[...,0,:,:])
    ldet = chol_logdet(V[...,0,:,:])
    for n in range(N-1):
        # Compute the solution of the system
        x[...,n+1,:] = (y[...,n+1,:]
                        - mvdot(misc.T(B[...,n,:,:]),
                                chol_solve(V[...,n,:,:],
                                           x[...,n,:])))
        # Compute the superdiagonal block of the inverse
        C[...,n,:,:] = chol_solve(V[...,n,:,:],
                                  B[...,n,:,:],
                                  matrix=True)
        # Compute the diagonal block
        V[...,n+1,:,:] = (A[...,n+1,:,:]
                        - mmdot(misc.T(B[...,n,:,:]), C[...,n,:,:]))
        # Ensure symmetry by 0.5*(V+V.T)
        V[...,n+1,:,:] = 0.5 * (V[...,n+1,:,:] + misc.T(V[...,n+1,:,:]))
        # Compute and store the Cholesky factor of the diagonal block
        V[...,n+1,:,:] = chol(V[...,n+1,:,:])
        # Compute the log-det term here, too
        ldet += chol_logdet(V[...,n+1,:,:])

    #
    # Backward recursion
    #
    x[...,-1,:] = chol_solve(V[...,-1,:,:], x[...,-1,:])
    V[...,-1,:,:] = chol_inv(V[...,-1,:,:])
    for n in reversed(range(N-1)):
        # Compute the solution of the system
        x[...,n,:] = chol_solve(V[...,n,:,:],
                                x[...,n,:] - mvdot(B[...,n,:,:],
                                                   x[...,n+1,:]))
        # Compute the diagonal block of the inverse
        V[...,n,:,:] = (chol_inv(V[...,n,:,:])
                        + mmdot(C[...,n,:,:],
                                mmdot(V[...,n+1,:,:],
                                misc.T(C[...,n,:,:]))))
        C[...,n,:,:] = - mmdot(C[...,n,:,:], V[...,n+1,:,:])
        # Ensure symmetry by 0.5*(V+V.T)
        V[...,n,:,:] = 0.5 * (V[...,n,:,:] + misc.T(V[...,n,:,:]))

    return (V, C, x, ldet)



################################################################################
# Copyright (C) 2014 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


"""
Unit tests for `transformations` module.
"""

import numpy as np

from bayespy.inference.vmp.nodes.gaussian import GaussianARD
from bayespy.inference.vmp.nodes.gaussian import Gaussian
from bayespy.inference.vmp.nodes.gamma import Gamma
from bayespy.inference.vmp.nodes.wishart import Wishart
from bayespy.inference.vmp.nodes.dot import SumMultiply
from bayespy.inference.vmp.nodes.gaussian_markov_chain import GaussianMarkovChain

from bayespy.utils import linalg
from bayespy.utils import random
from bayespy.utils import optimize

from ..transformations import RotateGaussianARD
from ..transformations import RotateGaussianMarkovChain
from ..transformations import RotateVaryingMarkovChain

from bayespy.utils.misc import TestCase

class TestRotateGaussianARD(TestCase):

    def test_cost_function(self):
        """
        Test the speed-up rotation of Gaussian ARD arrays.
        """

        # Use seed for deterministic testing
        np.random.seed(42)

        def test(shape, plates, 
                 axis=-1, 
                 alpha_plates=None, 
                 plate_axis=None,
                 mu=3):

            if plate_axis is not None:
                precomputes = [False, True]
            else:
                precomputes = [False]
                
            for precompute in precomputes:
                # Construct the model
                D = shape[axis]
                if alpha_plates is not None:
                    alpha = Gamma(2, 2,
                                  plates=alpha_plates)
                    alpha.initialize_from_random()
                else:
                    alpha = 2
                X = GaussianARD(mu, alpha,
                                shape=shape,
                                plates=plates)

                # Some initial learning and rotator constructing
                X.initialize_from_random()
                Y = GaussianARD(X, 1)
                Y.observe(np.random.randn(*(Y.get_shape(0))))
                X.update()
                if alpha_plates is not None:
                    alpha.update()
                    true_cost0_alpha = alpha.lower_bound_contribution()
                    rotX = RotateGaussianARD(X, alpha, 
                                             axis=axis,
                                             precompute=precompute)
                else:
                    rotX = RotateGaussianARD(X, 
                                             axis=axis,
                                             precompute=precompute)
                true_cost0_X = X.lower_bound_contribution()

                # Rotation matrices
                I = np.identity(D)
                R = np.random.randn(D, D)
                if plate_axis is not None:
                    C = plates[plate_axis]
                    Q = np.random.randn(C, C)
                    Ic = np.identity(C)
                else:
                    Q = None
                    Ic = None

                # Compute bound terms
                rotX.setup(plate_axis=plate_axis)
                rot_cost0 = rotX.get_bound_terms(I, Q=Ic)
                rot_cost1 = rotX.get_bound_terms(R, Q=Q)
                self.assertAllClose(sum(rot_cost0.values()),
                                    rotX.bound(I, Q=Ic)[0],
                                    msg="Bound terms and total bound differ")
                self.assertAllClose(sum(rot_cost1.values()),
                                    rotX.bound(R, Q=Q)[0],
                                    msg="Bound terms and total bound differ")
                # Perform rotation
                rotX.rotate(R, Q=Q)
                # Check bound terms
                true_cost1_X = X.lower_bound_contribution()
                self.assertAllClose(true_cost1_X - true_cost0_X,
                                    rot_cost1[X] - rot_cost0[X],
                                    msg="Incorrect rotation cost for X")
                if alpha_plates is not None:
                    true_cost1_alpha = alpha.lower_bound_contribution()
                    self.assertAllClose(true_cost1_alpha - true_cost0_alpha,
                                        rot_cost1[alpha] - rot_cost0[alpha],
                                        msg="Incorrect rotation cost for alpha")
            return

        # Rotating a vector (zero mu)
        test( (3,), (),    axis=-1,                       mu=0)
        test( (3,), (),    axis=-1, alpha_plates=(1,),    mu=0)
        test( (3,), (),    axis=-1, alpha_plates=(3,),    mu=0)
        test( (3,), (2,4), axis=-1,                       mu=0)
        test( (3,), (2,4), axis=-1, alpha_plates=(1,),    mu=0)
        test( (3,), (2,4), axis=-1, alpha_plates=(3,),    mu=0)
        test( (3,), (2,4), axis=-1, alpha_plates=(2,4,3), mu=0)
        test( (3,), (2,4), axis=-1, alpha_plates=(1,4,3), mu=0)

        # Rotating a vector (full mu)
        test( (3,), (),    axis=-1,                       mu=3*np.ones((3,)))
        test( (3,), (),    axis=-1, alpha_plates=(),      mu=3*np.ones((3,)))
        test( (3,), (),    axis=-1, alpha_plates=(1,),    mu=3*np.ones((3,)))
        test( (3,), (),    axis=-1, alpha_plates=(3,),    mu=3*np.ones((3,)))
        test( (3,), (2,4), axis=-1,                       mu=3*np.ones((2,4,3)))
        test( (3,), (2,4), axis=-1, alpha_plates=(1,),    mu=3*np.ones((2,4,3)))
        test( (3,), (2,4), axis=-1, alpha_plates=(3,),    mu=3*np.ones((2,4,3)))
        test( (3,), (2,4), axis=-1, alpha_plates=(2,4,3), mu=3*np.ones((2,4,3)))
        test( (3,), (2,4), axis=-1, alpha_plates=(1,4,3), mu=3*np.ones((2,4,3)))

        # Rotating a vector (broadcast mu)
        test( (3,), (),      axis=-1,                         mu=3*np.ones((1,)))
        test( (3,), (),      axis=-1, alpha_plates=(1,),      mu=3*np.ones((1,)))
        test( (3,), (),      axis=-1, alpha_plates=(3,),      mu=3*np.ones((1,)))
        test( (3,), (2,4,5), axis=-1,                         mu=3*np.ones((4,1,1)))
        test( (3,), (2,4,5), axis=-1, alpha_plates=(1,),      mu=3*np.ones((4,1,1)))
        test( (3,), (2,4,5), axis=-1, alpha_plates=(3,),      mu=3*np.ones((4,1,1)))
        test( (3,), (2,4,5), axis=-1, alpha_plates=(2,4,5,3), mu=3*np.ones((4,1,1))) #!!
        test( (3,), (2,4,5), axis=-1, alpha_plates=(1,4,5,3), mu=3*np.ones((4,1,1)))

        # Rotating an array
        test( (2,3,4), (),    axis=-1)
        test( (2,3,4), (),    axis=-2)
        test( (2,3,4), (),    axis=-3)
        test( (2,3,4), (5,6), axis=-1)
        test( (2,3,4), (5,6), axis=-2)
        test( (2,3,4), (5,6), axis=-3)
        test( (2,3,4), (5,6), axis=-1, alpha_plates=(3,1))
        test( (2,3,4), (5,6), axis=-2, alpha_plates=(3,1))
        test( (2,3,4), (5,6), axis=-3, alpha_plates=(3,1))
        test( (2,3), (4,5,6), axis=-1, alpha_plates=(5,1,2,1))
        test( (2,3), (4,5,6), axis=-2, alpha_plates=(5,1,2,1))

        # Test mu array broadcasting
        test( (2,3,4), (5,6,7), axis=-2,
              mu=GaussianARD(3, 1,
                             shape=(3,1),
                             plates=(6,1,1)))
        test( (2,3,4), (5,6,7), axis=-3,
              mu=GaussianARD(3, 1,
                             shape=(3,1),
                             plates=(6,1,1)))
        test( (2,3,4), (5,6,7), axis=-2, alpha_plates=(5,1,7,2,1,1),
              mu=GaussianARD(3, 1,
                             shape=(3,1),
                             plates=(6,1,1)))

        # Plate rotation
        test( (3,), (5,), axis=-1, plate_axis=-1)
        test( (3,), (4,5,6), axis=-1, plate_axis=-2)
        test( (2,3), (4,5,6), axis=-2, plate_axis=-2)
        test( (2,3,4), (5,6,7), axis=-2, plate_axis=-3)

        # Plate rotation with alpha
        test( (2,3,4), (5,6,7), axis=-2, alpha_plates=(3,1), plate_axis=-2)
        test( (2,3,4), (5,6,7), axis=-2, alpha_plates=(6,1,2,1,4), plate_axis=-3)

        # Plate rotation with mu
        test( (2,3,4), (5,6,7), axis=-2, plate_axis=-2,
              mu=GaussianARD(3, 1,
                             shape=(3,1),
                             plates=(6,1,1)))
        test( (2,3,4), (5,6,7), axis=-3, plate_axis=-2,
              mu=GaussianARD(3, 1,
                             shape=(3,1),
                             plates=(6,1,1)))
        test( (2,3,4), (5,6,7), axis=-2, alpha_plates=(5,1,7,2,1,1), plate_axis=-2,
              mu=GaussianARD(3, 1,
                             shape=(3,1),
                             plates=(6,1,1)))

        #
        # Plate rotation with mu and alpha
        #

        # Basic, matching sizes
        test( (3,), (4,), axis=-1, plate_axis=-1,
              alpha_plates=(4,3),
              mu=GaussianARD(3, 1,
                             shape=(3,),
                             plates=(4,)))
        # Broadcast for mu
        test( (3,), (4,), axis=-1, plate_axis=-1,
              alpha_plates=(4,3),
              mu=GaussianARD(3, 1,
                             shape=(1,),
                             plates=(4,)))
        test( (3,), (4,), axis=-1, plate_axis=-1,
              alpha_plates=(4,3),
              mu=GaussianARD(3, 1,
                             shape=(),
                             plates=(1,)))
        test( (3,), (4,), axis=-1, plate_axis=-1,
              alpha_plates=(4,3),
              mu=GaussianARD(3, 1,
                             shape=(3,),
                             plates=(1,)))
        # Broadcast for alpha
        test( (3,), (4,), axis=-1, plate_axis=-1,
              alpha_plates=(4,1),
              mu=GaussianARD(3, 1,
                             shape=(3,),
                             plates=(4,)))
        test( (3,), (4,), axis=-1, plate_axis=-1,
              alpha_plates=(3,),
              mu=GaussianARD(3, 1,
                             shape=(3,),
                             plates=(4,)))
        # Several variable dimensions
        test( (3,4,5), (2,), axis=-2, plate_axis=-1,
              alpha_plates=(2,3,4,5),
              mu=GaussianARD(3, 1,
                             shape=(3,4,5),
                             plates=(2,)))
        test( (3,4,5), (2,), axis=-2, plate_axis=-1,
              alpha_plates=(2,3,1,5),
              mu=GaussianARD(3, 1,
                             shape=(4,1),
                             plates=(2,1)))
        # Several plate dimensions
        test( (5,), (2,3,4), axis=-1, plate_axis=-2,
              alpha_plates=(2,3,4,5),
              mu=GaussianARD(3, 1,
                             shape=(5,),
                             plates=(2,3,4)))
        # Several plate dimensions, rotated plate broadcasted in alpha
        test( (5,), (2,3,4), axis=-1, plate_axis=-2,
              alpha_plates=(2,1,4,5),
              mu=GaussianARD(3, 1,
                             shape=(5,),
                             plates=(2,3,4)))
        test( (5,), (2,3,4), axis=-1, plate_axis=-2,
              alpha_plates=(4,5),
              mu=GaussianARD(3, 1,
                             shape=(5,),
                             plates=(2,3,4)))
        # Several plate dimensions, rotated plate broadcasted in mu
        test( (5,), (2,3,4), axis=-1, plate_axis=-2,
              alpha_plates=(2,3,4,5),
              mu=GaussianARD(3, 1,
                             shape=(5,),
                             plates=(2,1,4)))
        test( (5,), (2,3,4), axis=-1, plate_axis=-2,
              alpha_plates=(2,3,4,5),
              mu=GaussianARD(3, 1,
                             shape=(5,),
                             plates=(4,)))
        # Several plate dimensions, rotated plate broadcasted in mu and alpha
        test( (5,), (2,3,4), axis=-1, plate_axis=-2,
              alpha_plates=(2,1,4,5),
              mu=GaussianARD(3, 1,
                             shape=(5,),
                             plates=(2,1,4)))
        test( (5,), (2,3,4), axis=-1, plate_axis=-2,
              alpha_plates=(4,5),
              mu=GaussianARD(3, 1,
                             shape=(5,),
                             plates=(4,)))

        # TODO: Missing values
        
        pass

    def test_cost_gradient(self):
        """
        Test gradient of the rotation cost function for Gaussian ARD arrays.
        """

        # Use seed for deterministic testing
        np.random.seed(42)

        def test(shape, plates, 
                 axis=-1, 
                 alpha_plates=None, 
                 plate_axis=None,
                 mu=3):
            
            if plate_axis is not None:
                precomputes = [False, True]
            else:
                precomputes = [False]
                
            for precompute in precomputes:
                # Construct the model
                D = shape[axis]
                if alpha_plates is not None:
                    alpha = Gamma(3, 5,
                                  plates=alpha_plates)
                    alpha.initialize_from_random()
                else:
                    alpha = 2
                X = GaussianARD(mu, alpha,
                                shape=shape,
                                plates=plates)

                # Some initial learning and rotator constructing
                X.initialize_from_random()
                Y = GaussianARD(X, 1)
                Y.observe(np.random.randn(*(Y.get_shape(0))))
                X.update()
                if alpha_plates is not None:
                    alpha.update()
                    rotX = RotateGaussianARD(X, alpha, 
                                             axis=axis,
                                             precompute=precompute)
                else:
                    rotX = RotateGaussianARD(X, 
                                             axis=axis,
                                             precompute=precompute)
                try:
                    mu.update()
                except:
                    pass

                # Rotation matrices
                R = np.random.randn(D, D)
                if plate_axis is not None:
                    C = plates[plate_axis]
                    Q = np.random.randn(C, C)
                else:
                    Q = None

                # Compute bound terms
                rotX.setup(plate_axis=plate_axis)

                if plate_axis is None:
                    def f_r(r):
                        (b, dr) = rotX.bound(np.reshape(r, np.shape(R)))
                        return (b, np.ravel(dr))
                else:
                    def f_r(r):
                        (b, dr, dq) = rotX.bound(np.reshape(r, np.shape(R)),
                                             Q=Q)
                        return (b, np.ravel(dr))

                    def f_q(q):
                        (b, dr, dq) = rotX.bound(R,
                                             Q=np.reshape(q, np.shape(Q)))
                        return (b, np.ravel(dq))

                # Check gradient with respect to R
                err = optimize.check_gradient(f_r, 
                                              np.ravel(R), 
                                              verbose=False)[1]
                self.assertAllClose(err, 0, 
                                    atol=1e-4,
                                    msg="Gradient incorrect for R")

                # Check gradient with respect to Q
                if plate_axis is not None:
                    err = optimize.check_gradient(f_q, 
                                                  np.ravel(Q), 
                                                  verbose=False)[1]
                    self.assertAllClose(err, 0,
                                        atol=1e-4,
                                        msg="Gradient incorrect for Q")

            return

        #
        # Basic rotation
        #
        test((3,), (), axis=-1)
        test((2,3,4), (), axis=-1)
        test((2,3,4), (), axis=-2)
        test((2,3,4), (), axis=-3)
        test((2,3,4), (5,6), axis=-2)

        #
        # Rotation with mu
        #

        # Simple
        test((1,), (), axis=-1,
             mu=GaussianARD(2, 4,
                            shape=(1,),
                            plates=()))
        test((3,), (), axis=-1,
             mu=GaussianARD(2, 4,
                            shape=(3,),
                            plates=()))
        # Broadcast mu over rotated dim
        test((3,), (), axis=-1,
             mu=GaussianARD(2, 4,
                            shape=(1,),
                            plates=()))
        test((3,), (), axis=-1,
             mu=GaussianARD(2, 4,
                            shape=(),
                            plates=()))
        # Broadcast mu over dim when multiple dims
        test((2,3), (), axis=-1,
             mu=GaussianARD(2, 4,
                            shape=(1,3),
                            plates=()))
        test((2,3), (), axis=-1,
             mu=GaussianARD(2, 4,
                            shape=(3,),
                            plates=()))
        # Broadcast mu over rotated dim when multiple dims
        test((2,3), (), axis=-2,
             mu=GaussianARD(2, 4,
                            shape=(1,3),
                            plates=()))
        test((2,3), (), axis=-2,
             mu=GaussianARD(2, 4,
                            shape=(3,),
                            plates=()))
        # Broadcast mu over plates
        test((3,), (4,5), axis=-1,
             mu=GaussianARD(2, 4,
                            shape=(3,),
                            plates=(4,1)))
        test((3,), (4,5), axis=-1,
             mu=GaussianARD(2, 4,
                            shape=(3,),
                            plates=(5,)))

        #
        # Rotation with alpha
        #

        # Simple
        test((1,), (), axis=-1,
             alpha_plates=())
        test((3,), (), axis=-1,
             alpha_plates=(3,))
        # Broadcast alpha over rotated dim
        test((3,), (), axis=-1,
             alpha_plates=())
        test((3,), (), axis=-1,
             alpha_plates=(1,))
        # Broadcast alpha over dim when multiple dims
        test((2,3), (), axis=-1,
             alpha_plates=(1,3))
        test((2,3), (), axis=-1,
             alpha_plates=(3,))
        # Broadcast alpha over rotated dim when multiple dims
        test((2,3), (), axis=-2,
             alpha_plates=(1,3))
        test((2,3), (), axis=-2,
             alpha_plates=(3,))
        # Broadcast alpha over plates
        test((3,), (4,5), axis=-1,
             alpha_plates=(4,1,3))
        test((3,), (4,5), axis=-1,
             alpha_plates=(5,3))

        #
        # Rotation with alpha and mu
        #

        # Simple
        test((1,), (), axis=-1,
             alpha_plates=(1,),
             mu=GaussianARD(2, 4,
                            shape=(1,),
                            plates=()))
        test((3,), (), axis=-1,
             alpha_plates=(3,),
             mu=GaussianARD(2, 4,
                            shape=(3,),
                            plates=()))
        # Broadcast mu over rotated dim
        test((3,), (), axis=-1,
             alpha_plates=(3,),
             mu=GaussianARD(2, 4,
                            shape=(1,),
                            plates=()))
        test((3,), (), axis=-1,
             alpha_plates=(3,),
             mu=GaussianARD(2, 4,
                            shape=(),
                            plates=()))
        # Broadcast alpha over rotated dim
        test((3,), (), axis=-1,
             alpha_plates=(1,),
             mu=GaussianARD(2, 4,
                            shape=(3,),
                            plates=()))
        test((3,), (), axis=-1,
             alpha_plates=(),
             mu=GaussianARD(2, 4,
                            shape=(3,),
                            plates=()))
        # Broadcast both mu and alpha over rotated dim
        test((3,), (), axis=-1,
             alpha_plates=(1,),
             mu=GaussianARD(2, 4,
                            shape=(1,),
                            plates=()))
        test((3,), (), axis=-1,
             alpha_plates=(),
             mu=GaussianARD(2, 4,
                            shape=(),
                            plates=()))
        # Broadcast mu over plates
        test((3,), (4,5), axis=-1,
             alpha_plates=(4,5,3),
             mu=GaussianARD(2, 4,
                            shape=(3,),
                            plates=(4,1)))
        test((3,), (4,5), axis=-1,
             alpha_plates=(4,5,3),
             mu=GaussianARD(2, 4,
                            shape=(3,),
                            plates=(5,)))
        # Broadcast alpha over plates
        test((3,), (4,5), axis=-1,
             alpha_plates=(4,1,3),
             mu=GaussianARD(2, 4,
                            shape=(3,),
                            plates=(4,5)))
        test((3,), (4,5), axis=-1,
             alpha_plates=(5,3),
             mu=GaussianARD(2, 4,
                            shape=(3,),
                            plates=(4,5)))
        # Broadcast both mu and alpha over plates
        test((3,), (4,5), axis=-1,
             alpha_plates=(4,1,3),
             mu=GaussianARD(2, 4,
                            shape=(3,),
                            plates=(4,1)))
        test((3,), (4,5), axis=-1,
             alpha_plates=(5,3),
             mu=GaussianARD(2, 4,
                            shape=(3,),
                            plates=(5,)))
        # Broadcast both mu and alpha over plates but different plates
        test((3,), (4,5), axis=-1,
             alpha_plates=(4,1,3),
             mu=GaussianARD(2, 4,
                            shape=(3,),
                            plates=(5,)))
        test((3,), (4,5), axis=-1,
             alpha_plates=(5,3),
             mu=GaussianARD(2, 4,
                            shape=(3,),
                            plates=(4,1)))

        #
        # Rotation with missing values
        #

        # TODO

        #
        # Plate rotation
        #

        # Simple
        test((2,), (3,),    axis=-1, plate_axis=-1)
        test((2,), (3,4,5), axis=-1, plate_axis=-1)
        test((2,), (3,4,5), axis=-1, plate_axis=-2)
        test((2,), (3,4,5), axis=-1, plate_axis=-3)
        test((2,3), (4,5),  axis=-2, plate_axis=-2)

        # With mu
        test((2,), (3,), axis=-1, plate_axis=-1,
             mu=GaussianARD(3, 4,
                            shape=(2,),
                            plates=(3,)))
        # With mu broadcasted
        test((2,), (3,), axis=-1, plate_axis=-1,
             mu=GaussianARD(3, 4,
                            shape=(2,),
                            plates=(1,)))
        test((2,), (3,), axis=-1, plate_axis=-1,
             mu=GaussianARD(3, 4,
                            shape=(2,),
                            plates=()))
        # With mu multiple plates
        test((2,), (3,4,5), axis=-1, plate_axis=-2,
             mu=GaussianARD(3, 4,
                            shape=(2,),
                            plates=(3,4,5)))
        # With mu multiple dims
        test((2,3,4), (5,), axis=-2, plate_axis=-1,
             mu=GaussianARD(3, 4,
                            shape=(2,3,4),
                            plates=(5,)))

        #
        # With alpha
        #
        print("Test: Plate rotation with alpha. Scalars.")
        test((1,), (1,), axis=-1, plate_axis=-1,
             alpha_plates=(1,1),
             mu=0)
        print("Test: Plate rotation with alpha. Plates.")
        test((1,), (3,), axis=-1, plate_axis=-1,
             alpha_plates=(3,1),
             mu=0)
        print("Test: Plate rotation with alpha. Dims.")
        test((3,), (1,), axis=-1, plate_axis=-1,
             alpha_plates=(1,3),
             mu=0)
        print("Test: Plate rotation with alpha. Broadcast alpha over rotated plates.")
        test((1,), (3,), axis=-1, plate_axis=-1,
             alpha_plates=(1,1),
             mu=0)
        test((1,), (3,), axis=-1, plate_axis=-1,
             alpha_plates=(1,),
             mu=0)
        print("Test: Plate rotation with alpha. Broadcast alpha over dims.")
        test((3,), (1,), axis=-1, plate_axis=-1,
             alpha_plates=(1,1),
             mu=0)
        test((3,), (1,), axis=-1, plate_axis=-1,
             alpha_plates=(),
             mu=0)
        print("Test: Plate rotation with alpha. Multiple dims.")
        test((2,3,4,5), (6,), axis=-2, plate_axis=-1,
             alpha_plates=(6,2,3,4,5),
             mu=0)
        print("Test: Plate rotation with alpha. Multiple plates.")
        test((2,), (3,4,5), axis=-1, plate_axis=-1,
             alpha_plates=(3,4,5,2),
             mu=0)
        test((2,), (3,4,5), axis=-1, plate_axis=-2,
             alpha_plates=(3,4,5,2),
             mu=0)
        test((2,), (3,4,5), axis=-1, plate_axis=-3,
             alpha_plates=(3,4,5,2),
             mu=0)

        #
        # With alpha and mu
        #
        print("Test: Plate rotation with alpha and mu. Scalars.")
        test((1,), (1,), axis=-1, plate_axis=-1,
             alpha_plates=(1,1),
             mu=GaussianARD(2, 3,
                            shape=(1,),
                            plates=(1,)))
        print("Test: Plate rotation with alpha and mu. Plates.")
        test((1,), (3,), axis=-1, plate_axis=-1,
             alpha_plates=(3,1),
             mu=GaussianARD(2, 3,
                            shape=(1,),
                            plates=(3,)))
        print("Test: Plate rotation with alpha and mu. Dims.")
        test((3,), (1,), axis=-1, plate_axis=-1,
             alpha_plates=(1,3),
             mu=GaussianARD(2, 3,
                            shape=(3,),
                            plates=(1,)))
        print("Test: Plate rotation with alpha and mu. Broadcast over rotated "
              "plates.")
        test((1,), (3,), axis=-1, plate_axis=-1,
             alpha_plates=(1,1),
             mu=GaussianARD(2, 3,
                            shape=(1,),
                            plates=(1,)))
        test((1,), (3,), axis=-1, plate_axis=-1,
             alpha_plates=(1,),
             mu=GaussianARD(2, 3,
                            shape=(1,),
                            plates=()))
        print("Test: Plate rotation with alpha and mu. Broadcast over dims.")
        test((3,), (1,), axis=-1, plate_axis=-1,
             alpha_plates=(1,1),
             mu=GaussianARD(2, 3,
                            shape=(1,),
                            plates=(1,)))
        test((3,), (1,), axis=-1, plate_axis=-1,
             alpha_plates=(),
             mu=GaussianARD(2, 3,
                            shape=(),
                            plates=(1,)))
        print("Test: Plate rotation with alpha and mu. Multiple dims.")
        test((2,3,4,5), (6,), axis=-2, plate_axis=-1,
             alpha_plates=(6,2,3,4,5),
             mu=GaussianARD(2, 3,
                            shape=(2,3,4,5),
                            plates=(6,)))
        print("Test: Plate rotation with alpha and mu. Multiple plates.")
        test((2,), (3,4,5), axis=-1, plate_axis=-1,
             alpha_plates=(3,4,5,2),
             mu=GaussianARD(2, 3,
                            shape=(2,),
                            plates=(3,4,5,)))
        test((2,), (3,4,5), axis=-1, plate_axis=-2,
             alpha_plates=(3,4,5,2),
             mu=GaussianARD(2, 3,
                            shape=(2,),
                            plates=(3,4,5,)))
        test((2,), (3,4,5), axis=-1, plate_axis=-3,
             alpha_plates=(3,4,5,2),
             mu=GaussianARD(2, 3,
                            shape=(2,),
                            plates=(3,4,5,)))

        # TODO: With missing values
        
        pass

class TestRotateGaussianMarkovChain(TestCase):

    def test_cost_function(self):
        """
        Test the cost function of the speed-up rotation for Markov chain
        """

        np.random.seed(42)

        def check(D, N, mu=None, Lambda=None, rho=None, A=None):
            if mu is None:
                mu = np.zeros(D)
            if Lambda is None:
                Lambda = np.identity(D)
            if rho is None:
                rho = np.ones(D)
            if A is None:
                A = GaussianARD(3, 5,
                                shape=(D,),
                                plates=(D,))
                
            V = np.identity(D) + np.ones((D,D))

            # Construct model
            X = GaussianMarkovChain(mu,
                                    Lambda,
                                    A,
                                    rho,
                                    n=N+1,
                                    initialize=False)
            Y = Gaussian(X,
                         V,
                         initialize=False)

            # Posterior estimation
            Y.observe(np.random.randn(*(Y.get_shape(0))))
            X.update()
            try:
                A.update()
            except:
                pass
            try:
                mu.update()
            except:
                pass
            try:
                Lambda.update()
            except:
                pass
            try:
                rho.update()
            except:
                pass

            # Construct rotator
            rotA = RotateGaussianARD(A, axis=-1)
            rotX = RotateGaussianMarkovChain(X, rotA)

            # Rotation
            true_cost0 = X.lower_bound_contribution()
            rotX.setup()
            I = np.identity(D)
            R = np.random.randn(D, D)
            rot_cost0 = rotX.get_bound_terms(I)
            rot_cost1 = rotX.get_bound_terms(R)
            self.assertAllClose(sum(rot_cost0.values()),
                                rotX.bound(I)[0],
                                    msg="Bound terms and total bound differ")
            self.assertAllClose(sum(rot_cost1.values()),
                                rotX.bound(R)[0],
                                msg="Bound terms and total bound differ")
            rotX.rotate(R)
            true_cost1 = X.lower_bound_contribution()
            self.assertAllClose(true_cost1 - true_cost0,
                                rot_cost1[X] - rot_cost0[X],
                                msg="Incorrect rotation cost for X")
            
            return

        self._run_checks(check)

        pass

    def test_cost_gradient(self):
        """
        Test the gradient of the speed-up rotation for Markov chain
        """

        # Use seed for deterministic testing
        np.random.seed(42)

        def check(D, N, mu=None, Lambda=None, rho=None, A=None):
            if mu is None:
                mu = np.zeros(D)
            if Lambda is None:
                Lambda = np.identity(D)
            if rho is None:
                rho = np.ones(D)
            if A is None:
                A = GaussianARD(3, 5,
                                shape=(D,),
                                plates=(D,))
                
            V = np.identity(D) + np.ones((D,D))

            # Construct model
            X = GaussianMarkovChain(mu,
                                    Lambda,
                                    A,
                                    rho,
                                    n=N+1,
                                    initialize=False)
            Y = Gaussian(X,
                         V,
                         initialize=False)

            # Posterior estimation
            Y.observe(np.random.randn(*(Y.get_shape(0))))
            X.update()
            try:
                A.update()
            except:
                pass
            try:
                mu.update()
            except:
                pass
            try:
                Lambda.update()
            except:
                pass
            try:
                rho.update()
            except:
                pass

            # Construct rotator
            rotA = RotateGaussianARD(A, axis=-1)
            rotX = RotateGaussianMarkovChain(X, rotA)
            rotX.setup()

            # Check gradient with respect to R
            R = np.random.randn(D, D)
            def cost(r):
                (b, dr) = rotX.bound(np.reshape(r, np.shape(R)))
                return (b, np.ravel(dr))

            err = optimize.check_gradient(cost, 
                                          np.ravel(R), 
                                          verbose=False)[1]
            self.assertAllClose(err, 0, 
                                atol=1e-5,
                                msg="Gradient incorrect")
            
            return

        self._run_checks(check)

        pass

    def _run_checks(self, check):
        
        # Basic test
        check(2, 3)

        # Test mu
        check(2, 3,
              mu=GaussianARD(2, 4,
                             shape=(2,),
                             plates=()))
        check(2, 3,
              mu=GaussianARD(2, 4,
                             shape=(2,),
                             plates=(5,)))

        # Test Lambda
        check(2, 3,
              Lambda=Wishart(3, random.covariance(2)))
        check(2, 3,
              Lambda=Wishart(3, random.covariance(2),
                             plates=(5,)))

        # Test A
        check(2, 3,
              A=GaussianARD(2, 4,
                            shape=(2,),
                            plates=(2,)))
        check(2, 3,
              A=GaussianARD(2, 4,
                            shape=(2,),
                            plates=(3,2)))
        check(2, 3,
              A=GaussianARD(2, 4,
                            shape=(2,),
                            plates=(5,3,2)))

        # Test Lambda and mu
        check(2, 3,
              mu=GaussianARD(2, 4,
                             shape=(2,),
                             plates=()),
              Lambda=Wishart(2, random.covariance(2)))
        check(2, 3,
              mu=GaussianARD(2, 4,
                             shape=(2,),
                             plates=(5,)),
              Lambda=Wishart(2, random.covariance(2),
                             plates=(5,)))

        # Test mu and A
        check(2, 3,
              mu=GaussianARD(2, 4,
                             shape=(2,),
                             plates=()),
              A=GaussianARD(2, 4,
                            shape=(2,),
                            plates=(2,)))
        check(2, 3,
              mu=GaussianARD(2, 4,
                             shape=(2,),
                             plates=(5,)),
              A=GaussianARD(2, 4,
                            shape=(2,),
                            plates=(5,1,2,)))

        # Test Lambda and A
        check(2, 3,
              Lambda=Wishart(2, random.covariance(2)),
              A=GaussianARD(2, 4,
                            shape=(2,),
                            plates=(2,)))
        check(2, 3,
              Lambda=Wishart(2, random.covariance(2),
                             plates=(5,)),
              A=GaussianARD(2, 4,
                            shape=(2,),
                            plates=(5,1,2,)))

        # Test mu, Lambda and A
        check(2, 3,
              mu=GaussianARD(2, 4,
                             shape=(2,),
                             plates=()),
              Lambda=Wishart(2, random.covariance(2)),
              A=GaussianARD(2, 4,
                            shape=(2,),
                            plates=(2,)))
        check(2, 3,
              mu=GaussianARD(2, 4,
                             shape=(2,),
                             plates=(5,)),
              Lambda=Wishart(2, random.covariance(2),
                             plates=(5,)),
              A=GaussianARD(2, 4,
                            shape=(2,),
                            plates=(5,1,2,)))

        pass


    
class TestRotateVaryingMarkovChain(TestCase):

    def test_cost_function(self):
        """
        Test the speed-up rotation of Markov chain with time-varying dynamics
        """

        # Use seed for deterministic testing
        np.random.seed(42)

        def check(D, N, K,
                  mu=None,
                  Lambda=None,
                  rho=None):

            if mu is None:
                mu = np.zeros(D)
            if Lambda is None:
                Lambda = np.identity(D)
            if rho is None:
                rho = np.ones(D)

            V = np.identity(D) + np.ones((D,D))

            # Construct model
            B = GaussianARD(3, 5,
                            shape=(D,K),
                            plates=(1,D))
            S = GaussianARD(2, 4,
                            shape=(K,),
                            plates=(N,1))
            A = SumMultiply('dk,k->d', B, S)
            X = GaussianMarkovChain(mu,
                                    Lambda,
                                    A,
                                    rho,
                                    n=N+1,
                                    initialize=False)
            Y = Gaussian(X,
                         V,
                         initialize=False)

            # Posterior estimation
            Y.observe(np.random.randn(N+1,D))
            X.update()
            B.update()
            S.update()
            try:
                mu.update()
            except:
                pass
            try:
                Lambda.update()
            except:
                pass
            try:
                rho.update()
            except:
                pass

            # Construct rotator
            rotB = RotateGaussianARD(B, axis=-2)
            rotX = RotateVaryingMarkovChain(X, B, S, rotB)

            # Rotation
            true_cost0 = X.lower_bound_contribution()
            rotX.setup()
            I = np.identity(D)
            R = np.random.randn(D, D)
            rot_cost0 = rotX.get_bound_terms(I)
            rot_cost1 = rotX.get_bound_terms(R)
            self.assertAllClose(sum(rot_cost0.values()),
                                rotX.bound(I)[0],
                                    msg="Bound terms and total bound differ")
            self.assertAllClose(sum(rot_cost1.values()),
                                rotX.bound(R)[0],
                                msg="Bound terms and total bound differ")
            rotX.rotate(R)
            true_cost1 = X.lower_bound_contribution()
            self.assertAllClose(true_cost1 - true_cost0,
                                rot_cost1[X] - rot_cost0[X],
                                msg="Incorrect rotation cost for X")
            
            return

        self._run_checks(check)
        
        pass


    def test_cost_gradient(self):
        """
        Test the gradient of the rotation for MC with time-varying dynamics
        """

        # Use seed for deterministic testing
        np.random.seed(42)

        def check(D, N, K,
                  mu=None,
                  Lambda=None,
                  rho=None):

            if mu is None:
                mu = np.zeros(D)
            if Lambda is None:
                Lambda = np.identity(D)
            if rho is None:
                rho = np.ones(D)

            V = np.identity(D) + np.ones((D,D))

            # Construct model
            B = GaussianARD(3, 5,
                            shape=(D,K),
                            plates=(1,D))
            S = GaussianARD(2, 4,
                            shape=(K,),
                            plates=(N,1))
            A = SumMultiply('dk,k->d', B, S)
            X = GaussianMarkovChain(mu,
                                    Lambda,
                                    A,
                                    rho,
                                    n=N+1,
                                    initialize=False)
            Y = Gaussian(X,
                         V,
                         initialize=False)

            # Posterior estimation
            Y.observe(np.random.randn(N+1,D))
            X.update()
            B.update()
            S.update()
            try:
                mu.update()
            except:
                pass
            try:
                Lambda.update()
            except:
                pass
            try:
                rho.update()
            except:
                pass

            # Construct rotator
            rotB = RotateGaussianARD(B, axis=-2)
            rotX = RotateVaryingMarkovChain(X, B, S, rotB)
            rotX.setup()

            # Check gradient with respect to R
            R = np.random.randn(D, D)
            def cost(r):
                (b, dr) = rotX.bound(np.reshape(r, np.shape(R)))
                return (b, np.ravel(dr))

            err = optimize.check_gradient(cost, 
                                          np.ravel(R), 
                                          verbose=False)[1]
            self.assertAllClose(err, 0, 
                                atol=1e-6,
                                msg="Gradient incorrect")
            
            return

        self._run_checks(check)
        
        pass

    def _run_checks(self, check):
        
        # Basic test
        check(1, 1, 1)
        check(2, 1, 1)
        check(1, 2, 1)
        check(1, 1, 2)
        check(3, 4, 2)

        # Test mu
        check(2, 3, 4,
              mu=GaussianARD(2, 4,
                             shape=(2,),
                             plates=()))

        # Test Lambda
        check(2, 3, 4,
              Lambda=Wishart(3, random.covariance(2)))

        # Test Lambda and mu
        check(2, 3, 4,
              mu=GaussianARD(2, 4,
                             shape=(2,),
                             plates=()),
              Lambda=Wishart(2, random.covariance(2)))

        # TODO: Test plates

        pass




################################################################################
# Copyright (C) 2011-2013 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


"""
Functions for plotting nodes.

Functions
=========

.. currentmodule:: bayespy.plot

.. autosummary::
   :toctree: generated/

   pdf
   contour
   plot
   hinton
   gaussian_mixture_2d
   ellipse_from_cov
   ellipse_from_precision

Plotters
========

.. autosummary::
   :toctree: generated/

   Plotter
   PDFPlotter
   ContourPlotter
   HintonPlotter
   FunctionPlotter
   GaussianTimeseriesPlotter
   CategoricalMarkovChainPlotter
"""


import os, sys

############################################################################
# A STUPID WORKAROUND FOR A MATPLOTLIB 1.4.0 BUG RELATED TO INTERACTIVE MODE
# See: https://github.com/matplotlib/matplotlib/issues/3505
import __main__
if hasattr(__main__, '__file__'):
    sys.ps1 = ('WORKAROUND FOR A BUG #3505 IN MATPLOTLIB.\n'
               'IF YOU SEE THIS MESSAGE, TRY MATPLOTLIB!=1.4.0.')
# This workaround does not work on Python shell, only on stand-alone scripts
# and IPython. A better solution: require MPL!=1.4.0.
#############################################################################

import numpy as np
import scipy.sparse as sp
import scipy
from scipy import special
import matplotlib.pyplot as plt
from matplotlib import animation
from matplotlib import colors
from matplotlib import patches
#from matplotlib.pyplot import *

from bayespy.inference.vmp.nodes.categorical import CategoricalMoments
from bayespy.inference.vmp.nodes.gaussian import (GaussianMoments,
                                                  GaussianWishartMoments)
from bayespy.inference.vmp.nodes.beta import BetaMoments
from bayespy.inference.vmp.nodes.beta import DirichletMoments
from bayespy.inference.vmp.nodes.bernoulli import BernoulliMoments
from bayespy.inference.vmp.nodes.categorical import CategoricalMoments
from bayespy.inference.vmp.nodes.gamma import GammaMoments
from bayespy.inference.vmp.nodes.node import Node, Moments

from bayespy.utils import (misc,
                           random,
                           linalg)


# Users can use pyplot via this module
import matplotlib
mpl = matplotlib
pyplot = plt


def interactive(function):
    """A decorator for forcing functions to use the interactive mode.

    Parameters
    ----------

    function : callable
        The function to be decorated
    """

    def new_function(*args, **kwargs):
        if mpl.is_interactive():
            was_interactive = True
        else:
            was_interactive = False
            mpl.interactive(True)

        retval = function(*args, **kwargs)

        if not was_interactive:
            mpl.interactive(False)

        return retval

    return new_function


def _subplots(plotfunc, *args, fig=None, kwargs=None):
    """Create a collection of subplots

    Each subplot is created with the same plotting function.

    Inputs are given as pairs:

    (x, 3), (y, 2), ...

    where x,y,... are the input arrays and 3,2,... are the ndim
    parameters.  The last ndim axes of each array are interpreted as a
    single element to the plotting function.

    All high-level plotting functions should wrap low-level plotting
    functions with this function in order to generate subplots for
    plates.
    """

    if kwargs is None:
        kwargs = {}

    if fig is None:
        fig = plt.gcf()

    # Parse shape and plates of each input array
    shapes = [np.shape(x)[-n:] if n > 0 else ()
              for (x,n) in args]
    plates = [np.shape(x)[:-n] if n > 0 else np.shape(x)
              for (x,n) in args]

    # Get the full grid shape of the subplots
    broadcasted_plates = misc.broadcasted_shape(*plates)

    # Subplot indexing layout
    M = int(np.prod(broadcasted_plates[-2::-2]))
    N = int(np.prod(broadcasted_plates[-1::-2]))
    strides_subplot = [np.prod(broadcasted_plates[(j+2)::2]) * N
                       if ((len(broadcasted_plates)-j) % 2) == 0 else
                       np.prod(broadcasted_plates[(j+2)::2])
                       for j in range(len(broadcasted_plates))]

    # Plot each subplot
    for ind in misc.nested_iterator(broadcasted_plates):

        # Get the list of inputs for this subplot
        broadcasted_args = []
        for n in range(len(args)):
            i = misc.safe_indices(ind, plates[n])
            broadcasted_args.append(args[n][0][i])

        # Plot the subplot using the given function
        ind_subplot = int(np.einsum('i,i', ind, strides_subplot))
        axes = fig.add_subplot(M, N, ind_subplot+1)
        plotfunc(*broadcasted_args, axes=axes, **kwargs)


def pdf(Z, x, *args, name=None, axes=None, fig=None, **kwargs):
    """
    Plot probability density function of a scalar variable.

    Parameters
    ----------

    Z : node or function
        Stochastic node or log pdf function

    x : array
        Grid points
    """

    # TODO: Make it possible to plot a plated variable using _subplots function.

    if axes is None and fig is None:
        axes = plt.gca()
    else:
        if fig is None:
            fig = plt.gcf()
        axes = fig.add_subplot(111)

    try:
        lpdf = Z.logpdf(x)
    except AttributeError:
        lpdf = Z(x)
    p = np.exp(lpdf)
    retval = axes.plot(x, p, *args, **kwargs)

    if name is None:
        try:
            name = Z.name
        except AttributeError:
            pass

    if name:
        axes.set_title(r'$q(%s)$' % (name))
        axes.set_xlabel(r'$%s$' % (name))

    return retval


def contour(Z, x, y, n=None, axes=None, fig=None, **kwargs):
    """
    Plot 2-D probability density function of a 2-D variable.

    Parameters
    ----------

    Z : node or function
        Stochastic node or log pdf function

    x : array
        Grid points on x axis

    y : array
        Grid points on y axis
    """

    # TODO: Make it possible to plot a plated variable using _subplots function.

    if axes is None and fig is None:
        axes = plt.gca()
    else:
        if fig is None:
            fig = plt.gcf()
        axes = fig.add_subplot(111)

    XY = misc.grid(x, y)
    try:
        lpdf = Z.logpdf(XY)
    except AttributeError:
        lpdf = Z(XY)
    p = np.exp(lpdf)
    shape = (np.size(x), np.size(y))
    X = np.reshape(XY[:,0], shape)
    Y = np.reshape(XY[:,1], shape)
    P = np.reshape(p, shape)
    if n is not None:
        levels = np.linspace(0, np.amax(P), num=n+2)[1:-1]
        return axes.contour(X, Y, P, levels, **kwargs)
    else:
        return axes.contour(X, Y, P, **kwargs)



def plot_gaussian_mc(X, scale=2, **kwargs):
    """
    Plot Gaussian Markov chain as a 1-D function

    Parameters
    ----------
    X : node
        Node with Gaussian Markov chain moments.
    """
    timeseries_gaussian(X, axis=-2, scale=scale, **kwargs)


def plot_bernoulli(X, axis=-1, scale=2, **kwargs):
    """
    Plot Bernoulli node as a 1-D function
    """
    X = X._ensure_moments(X, BernoulliMoments)
    u_X = X.get_moments()
    z = u_X[0]
    return _timeseries_mean_and_error(z, None, axis=axis, **kwargs)


def plot_gaussian(X, axis=-1, scale=2, **kwargs):
    """
    Plot Gaussian node as a 1-D function

    Parameters
    ----------
    X : node
        Node with Gaussian moments.
    axis : int
        The index of the time axis.
    """
    X = X._ensure_moments(X, GaussianMoments, ndim=0)
    u_X = X.get_moments()
    x = u_X[0]
    xx = misc.get_diag(u_X[1], ndim=len(X.dims[0]))
    std = scale * np.sqrt(xx - x**2)
    #std = scale * np.sqrt(np.einsum('...ii->...i', xx) - x**2)

    return _timeseries_mean_and_error(x, std, axis=axis, **kwargs)


def plot(Y, axis=-1, scale=2, center=False, **kwargs):
    """
    Plot a variable or an array as 1-D function with errorbars
    """
    if misc.is_numeric(Y):
        return _timeseries_mean_and_error(Y, None, axis=axis, center=center, **kwargs)

    if isinstance(Y, Node):

        # Try Bernoulli plotting
        try:
            Y = Y._ensure_moments(Y, BernoulliMoments)
        except BernoulliMoments.NoConverterError:
            pass
        else:
            return plot_bernoulli(Y, axis=axis, scale=scale, center=center, **kwargs)

        # Try Gaussian plotting
        try:
            Y = Y._ensure_moments(Y, GaussianMoments, ndim=0)
        except GaussianMoments.NoConverterError:
            pass
        else:
            return plot_gaussian(Y, axis=axis, scale=scale, center=center, **kwargs)

    (mu, var) = Y.get_mean_and_variance()
    std = np.sqrt(var)

    return _timeseries_mean_and_error(mu, std,
                                      axis=axis,
                                      scale=scale,
                                      center=center,
                                      **kwargs)


# Some backward compatibility
def timeseries_gaussian_mc(*args, center=True, **kwargs):
    return plot_gaussian_mc(*args, center=center, **kwargs)
def timeseries_gaussian(*args, center=True, **kwargs):
    return plot_gaussian(*args, center=center, **kwargs)
timeseries_normal = timeseries_gaussian
def timeseries(*args, center=True, **kwargs):
    return plot(*args, center=center, **kwargs)


def _timeseries_mean_and_error(y, std, *args, axis=-1, center=True, fig=None, axes=None, **kwargs):
    # TODO/FIXME: You must multiply by ones(plates) in order to plot
    # broadcasted plates properly

    if fig is None:
        fig = plt.gcf()

    y = np.atleast_1d(y)
    shape = list(np.shape(y))

    # Get and remove the length of the time axis
    T = shape.pop(axis)

    # Move time axis to first
    y = np.rollaxis(y, axis)
    if std is not None:
        std = np.rollaxis(std, axis)

    y = np.reshape(y, (T, -1))
    if std is not None:
        std = np.reshape(std, (T, -1))

    # Remove 1s
    shape = [s for s in shape if s > 1]

    # Calculate number of rows and columns
    shape = misc.multiply_shapes(shape, (1,1))
    if len(shape) > 2:
        raise Exception("Can plot only in 2 dimensions (rows and columns)")
    (M, N) = shape

    # Prefer plotting to rows
    if M == 1:
        M = N
        N = 1

    # Plot each timeseries
    if axes is None:
        ax0 = fig.add_subplot(M, N, 1)
    for i in range(M*N):
        if axes is None:
            if i > 0:
                # Share x axis between all subplots
                ax = fig.add_subplot(M, N, i+1, sharex=ax0)
            else:
                ax = ax0

            # Autoscale the axes to data and use tight y and x axes
            ax.autoscale(enable=True, tight=True)
            ax.set_ylim(auto=True)

            if i < (M-1)*N:
                # Remove x tick labels from other than the last row
                plt.setp(ax.get_xticklabels(), visible=False)

        else:
            ax = axes[i]

        if std is None:
            errorplot(y=y[:,i], axes=ax, **kwargs)
        else:
            if len(args) > 0:
                raise Exception("Can't handle extra arguments")
            errorplot(y=y[:,i], error=std[:,i], axes=ax, **kwargs)

        if center:
            # Center the zero level on y-axis
            ylim = ax.get_ylim()
            vmax = np.max(np.abs(ylim))
            ax.set_ylim([-vmax, vmax])

    if axes is None:
        # Remove height space between subplots
        fig.subplots_adjust(hspace=0)


def _blob(axes, x, y, area, colour):
    """
    Draws a square-shaped blob with the given area (< 1) at
    the given coordinates.
    """
    hs = np.sqrt(area) / 2
    xcorners = np.array([x - hs, x + hs, x + hs, x - hs])
    ycorners = np.array([y - hs, y - hs, y + hs, y + hs])
    axes.fill(xcorners, ycorners, colour, edgecolor=colour)

def _rectangle(axes, x, y, width, height, **kwargs):
    _x = x - width/2
    _y = y - height/2
    rectangle = plt.Rectangle((_x, _y),
                              width,
                              height,
                              **kwargs)
    axes.add_patch(rectangle)
    return


def gaussian_mixture_2d(X, alpha=None, scale=2, fill=False, axes=None, **kwargs):
    """
    Plot Gaussian mixture as ellipses in 2-D

    Parameters
    ----------

    X : Mixture node

    alpha : Dirichlet-like node (optional)
       Probabilities for the clusters

    scale : float (optional)
       Scale for the covariance ellipses (by default, 2)
    """

    if axes is None:
        axes = plt.gca()

    mu_Lambda = X._ensure_moments(X.parents[1], GaussianWishartMoments)

    (mu, _, Lambda, _) = mu_Lambda.get_moments()
    mu = np.linalg.solve(Lambda, mu)

    if len(mu_Lambda.plates) != 1:
        raise NotImplementedError("Not yet implemented for more plates")

    K = mu_Lambda.plates[0]

    width = np.zeros(K)
    height = np.zeros(K)
    angle = np.zeros(K)

    for k in range(K):
        m = mu[k]
        L = Lambda[k]
        (u, W) = scipy.linalg.eigh(L)
        u[0] = np.sqrt(1/u[0])
        u[1] = np.sqrt(1/u[1])
        width[k] = 2*u[0]
        height[k] = 2*u[1]
        angle[k] = np.arctan(W[0,1] / W[0,0])

    angle = 180 * angle / np.pi
    mode_height = 1 / (width * height)

    # Use cluster probabilities to adjust alpha channel
    if alpha is not None:
        # Compute the normalized probabilities in a numerically stable way
        logsum_p = misc.logsumexp(alpha.u[0], axis=-1, keepdims=True)
        logp = alpha.u[0] - logsum_p
        p = np.exp(logp)
        # Visibility is based on cluster mode peak height
        visibility = mode_height * p
        visibility /= np.amax(visibility)
    else:
        visibility = np.ones(K)

    for k in range(K):
        ell = mpl.patches.Ellipse(mu[k], scale*width[k], scale*height[k],
                                  angle=(180+angle[k]),
                                  fill=fill,
                                  alpha=visibility[k],
                                  **kwargs)
        axes.add_artist(ell)

    plt.axis('equal')

    # If observed, plot the data too
    if np.any(X.observed):
        mask = np.array(X.observed) * np.ones(X.plates, dtype=np.bool)
        y = X.u[0][mask]
        plt.plot(y[:,0], y[:,1], 'r.')

    return


def _hinton(W, error=None, vmax=None, square=False, axes=None):
    """
    Draws a Hinton diagram for visualizing a weight matrix.

    Temporarily disables matplotlib interactive mode if it is on,
    otherwise this takes forever.

    Originally copied from
    http://wiki.scipy.org/Cookbook/Matplotlib/HintonDiagrams
    """

    if axes is None:
        axes = plt.gca()

    W = misc.atleast_nd(W, 2)
    (height, width) = W.shape
    if not vmax:
        #vmax = 2**np.ceil(np.log(np.max(np.abs(W)))/np.log(2))
        if error is not None:
            vmax = np.max(np.abs(W) + error)
        else:
            vmax = np.max(np.abs(W))

    axes.fill(0.5+np.array([0,width,width,0]),
              0.5+np.array([0,0,height,height]),
              'gray')
    if square:
        axes.set_aspect('equal')
    axes.set_ylim(0.5, height+0.5)
    axes.set_xlim(0.5, width+0.5)
    axes.set_xticks([])
    axes.set_yticks([])
    axes.invert_yaxis()
    for x in range(width):
        for y in range(height):
            _x = x+1
            _y = y+1
            w = W[y,x]
            _w = np.abs(w)
            if w > 0:
                _c = 'white'
            else:
                _c = 'black'
            if error is not None:
                e = error[y,x]
                if e < 0:
                    print(e, _w, vmax)
                    raise Exception("BUG? Negative error")
                if _w + e > vmax:
                    print(e, _w, vmax)
                    raise Exception("BUG? Value+error greater than max")
                _rectangle(axes,
                           _x,
                           _y,
                           min(1, np.sqrt((_w+e)/vmax)),
                           min(1, np.sqrt((_w+e)/vmax)),
                           edgecolor=_c,
                           fill=False)
            _blob(axes, _x, _y, min(1, _w/vmax), _c)


def matrix(A, axes=None):

    if axes is None:
        axes = plt.gca()

    A = np.atleast_2d(A)
    vmax = np.max(np.abs(A))
    return  axes.imshow(A,
                        interpolation='nearest',
                        cmap='RdBu_r',
                        vmin=-vmax,
                        vmax=vmax)

def new_matrix(A, vmax=None):
    A = np.atleast_2d(A)
    if vmax is None:
        vmax = np.max(np.abs(A))

    (M, N) = np.shape(A)

    for i in range(M):
        for j in range(N):
            pass

def gaussian_hinton(X, rows=None, cols=None, scale=1, fig=None):
    """
    Plot the Hinton diagram of a Gaussian node
    """

    if fig is None:
        fig = plt.gcf()

    # Get mean and second moment
    X = X._ensure_moments(X, GaussianMoments, ndim=0)
    (x, xx) = X.get_moments()
    ndim = len(X.dims[0])
    shape = X.get_shape(0)
    size = len(X.get_shape(0))

    # Compute standard deviation
    xx = misc.get_diag(xx, ndim=ndim)
    std = np.sqrt(xx - x**2)

    # Force explicit elements when broadcasting
    x = x * np.ones(shape)
    std = std * np.ones(shape)

    if rows is None:
        rows = np.nan
    if cols is None:
        cols = np.nan

    # Preprocess the axes to 0,...,ndim
    if rows < 0:
        rows += size
    if cols < 0:
        cols += size
    if rows < 0 or rows >= size:
        raise ValueError("Row axis invalid")
    if cols < 0 or cols >= size:
        raise ValueError("Column axis invalid")

    # Remove non-row and non-column axes that have length 1
    squeezed_shape = list(shape)
    for i in reversed(range(len(shape))):
        if shape[i] == 1 and i != rows and i != cols:
            squeezed_shape.pop(i)
            if i < cols:
                cols -= 1
            if i < rows:
                rows -= 1
    x = np.reshape(x, squeezed_shape)
    std = np.reshape(std, squeezed_shape)

    if np.ndim(x) < 2:
        cols += 2 - np.ndim(x)
        rows += 2 - np.ndim(x)
        x = np.atleast_2d(x)
        std = np.atleast_2d(std)

    size = np.ndim(x)
    if np.isnan(cols):
        if rows != size - 1:
            cols = size - 1
        else:
            cols = size - 2
    if np.isnan(rows):
        if cols != size - 1:
            rows = size - 1
        else:
            rows = size - 2

    # Put the row and column axes to the end
    axes = [i for i in range(size) if i not in (rows, cols)] + [rows, cols]
    x = np.transpose(x, axes=axes)
    std = np.transpose(std, axes=axes)

    vmax = np.max(np.abs(x) + scale*std)

    if scale == 0:
        _subplots(_hinton, (x, 2), fig=fig, kwargs=dict(vmax=vmax))
    else:
        def plotfunc(z, e, **kwargs):
            return _hinton(z, error=e, **kwargs)

        _subplots(plotfunc, (x, 2), (scale*std, 2), fig=fig, kwargs=dict(vmax=vmax))


def _hinton_figure(x, rows=None, cols=None, fig=None, square=True):
    """
    Plot the Hinton diagram of a Gaussian node
    """

    scale = 0
    std = 0

    if fig is None:
        fig = plt.gcf()

    # Get mean and second moment
    shape = np.shape(x)
    size = np.ndim(x)

    if rows is None:
        rows = np.nan
    if cols is None:
        cols = np.nan

    # Preprocess the axes to 0,...,ndim
    if rows < 0:
        rows += size
    if cols < 0:
        cols += size
    if rows < 0 or rows >= size:
        raise ValueError("Row axis invalid")
    if cols < 0 or cols >= size:
        raise ValueError("Column axis invalid")

    # Remove non-row and non-column axes that have length 1
    squeezed_shape = list(shape)
    for i in reversed(range(len(shape))):
        if shape[i] == 1 and i != rows and i != cols:
            squeezed_shape.pop(i)
            if i < cols:
                cols -= 1
            if i < rows:
                rows -= 1
    x = np.reshape(x, squeezed_shape)

    size = np.ndim(x)
    if np.isnan(cols):
        if rows != size - 1:
            cols = size - 1
        else:
            cols = size - 2
    if np.isnan(rows):
        if cols != size - 1:
            rows = size - 1
        else:
            rows = size - 2

    # Put the row and column axes to the end
    if np.ndim(x) >= 2:
        axes = [i for i in range(size) if i not in (rows, cols)] + [rows, cols]
        x = np.transpose(x, axes=axes)
        #std = np.transpose(std, axes=axes)

    vmax = np.max(np.abs(x) + scale*std)

    kw = dict(vmax=vmax, square=square)
    if scale == 0:
        _subplots(_hinton, (x, 2), fig=fig, kwargs=kw)
    else:
        def plotfunc(z, e, **kwargs):
            return _hinton(z, error=e, **kwargs)

        _subplots(plotfunc, (x, 2), (scale*std, 2), fig=fig, kwargs=kw)


# For backwards compatibility:
gaussian_array = gaussian_hinton


def timeseries_categorical_mc(Z, fig=None):

    if fig is None:
        fig = plt.gcf()

    # Make sure that the node is categorical
    Z = Z._ensure_moments(Z, CategoricalMoments, categories=None)

    # Get expectations (and broadcast explicitly)
    z = Z._message_to_child()[0] * np.ones(Z.get_shape(0))

    # Compute the subplot layout
    z = misc.atleast_nd(z, 4)
    if np.ndim(z) != 4:
        raise ValueError("Can not plot arrays with over 4 axes")
    M = np.shape(z)[0]
    N = np.shape(z)[1]

    # Plot Hintons
    for i in range(M):
        for j in range(N):
            axes = fig.add_subplot(M, N, i*N+j+1)
            _hinton(z[i,j].T, vmax=1.0, square=False, axes=axes)


def gamma_hinton(alpha, square=True, **kwargs):
    """
    Plot a beta distributed random variable as a Hinton diagram
    """

    # Make sure that the node is beta
    alpha = alpha._ensure_moments(alpha, GammaMoments)

    # Compute exp( <log p> )
    x = alpha.get_moments()[0]

    # Explicit broadcasting
    x = x * np.ones(alpha.plates)

    # Plot Hinton diagram
    return _hinton_figure(x, square=square, **kwargs)


def beta_hinton(P, square=True):
    """
    Plot a beta distributed random variable as a Hinton diagram
    """

    # Make sure that the node is beta
    P = P._ensure_moments(P, BetaMoments)

    # Compute exp( <log p> )
    p = np.exp(P._message_to_child()[0][...,0])

    # Explicit broadcasting
    p = p * np.ones(P.plates)

    # Plot Hinton diagram
    return _hinton(p, vmax=1.0, square=square)


def dirichlet_hinton(P, square=True):
    """
    Plot a beta distributed random variable as a Hinton diagram
    """

    # Make sure that the node is beta
    P = P._ensure_moments(P, DirichletMoments)

    # Compute exp( <log p> )
    p = np.exp(P._message_to_child()[0])

    # Explicit broadcasting
    p = p * np.ones(P.plates+(1,))

    # Plot Hinton diagram
    return _hinton(p, vmax=1.0, square=square)


def bernoulli_hinton(Z, square=True):
    """
    Plot a Bernoulli distributed random variable as a Hinton diagram
    """

    # Make sure that the node is Bernoulli
    Z = Z._ensure_moments(Z, BernoulliMoments)

    # Get <Z>
    z = Z._message_to_child()[0]

    # Explicit broadcasting
    z = z * np.ones(Z.plates)

    # Plot Hinton diagram
    return _hinton(z, vmax=1.0, square=square)


def categorical_hinton(Z, square=True):
    """
    Plot a Bernoulli distributed random variable as a Hinton diagram
    """

    # Make sure that the node is Bernoulli
    Z = Z._ensure_moments(Z, CategoricalMoments, categories=None)

    # Get <Z>
    z = Z._message_to_child()[0]

    # Explicit broadcasting
    z = z * np.ones(Z.plates+(1,))

    # Plot Hinton diagram
    return _hinton(np.squeeze(z), vmax=1.0, square=square)


def hinton(X, **kwargs):
    r"""
    Plot the Hinton diagram of a node

    The keyword arguments depend on the node type.  For some node types, the
    diagram also shows uncertainty with non-filled rectangles.  Currently,
    beta-like, Gaussian-like and Dirichlet-like nodes are supported.

    Parameters
    ----------

    X : node

    """

    if hasattr(X, "_ensure_moments"):

        try:
            X = X._ensure_moments(X, GaussianMoments, ndim=0)
        except Moments.NoConverterError:
            pass
        else:
            return gaussian_hinton(X, **kwargs)

        try:
            X = X._ensure_moments(X, GammaMoments)
        except Moments.NoConverterError:
            pass
        else:
            return gamma_hinton(X, **kwargs)

        try:
            X = X._ensure_moments(X, BetaMoments)
        except Moments.NoConverterError:
            pass
        else:
            return beta_hinton(X, **kwargs)

        try:
            X = X._ensure_moments(X, DirichletMoments)
        except Moments.NoConverterError:
            pass
        else:
            return dirichlet_hinton(X, **kwargs)

        try:
            X = X._ensure_moments(X, BernoulliMoments)
        except Moments.NoConverterError:
            pass
        else:
            return bernoulli_hinton(X, **kwargs)

        try:
            X = X._ensure_moments(X, CategoricalMoments, categories=None)
        except Moments.NoConverterError:
            pass
        else:
            return categorical_hinton(X, **kwargs)

    return _hinton_figure(X, **kwargs)


class Plotter():
    r"""
    Wrapper for plotting functions and base class for node plotters

    The purpose of this class is to collect all the parameters needed by a
    plotting function and provide a callable interface which needs only the node
    as the input.

    Plotter instances are callable objects that plot a given node using a
    specified plotting function.

    Parameters
    ----------

    plotter : function

        Plotting function to use

    args : defined by the plotting function

        Additional inputs needed by the plotting function

    kwargs : defined by the plotting function

        Additional keyword arguments supported by the plotting function

    Examples
    --------

    First, create a gamma variable:

    >>> import numpy as np
    >>> from bayespy.nodes import Gamma
    >>> x = Gamma(4, 5)

    The probability density function can be plotted as:

    >>> import bayespy.plot as bpplt
    >>> bpplt.pdf(x, np.linspace(0.1, 10, num=100))         # doctest: +ELLIPSIS
    [<matplotlib.lines.Line2D object at 0x...>]

    However, this can be problematic when one needs to provide a
    plotting function for the inference engine as the inference engine
    gives only the node as input.  Thus, we need to create a simple
    plotter wrapper:

    >>> p = bpplt.Plotter(bpplt.pdf, np.linspace(0.1, 10, num=100))

    Now, this callable object ``p`` needs only the node as the input:

    >>> p(x)                                                # doctest: +ELLIPSIS
    [<matplotlib.lines.Line2D object at 0x...>]

    Thus, it can be given to the inference engine to use as a plotting function:

    >>> x = Gamma(4, 5, plotter=p)
    >>> x.plot()                                            # doctest: +ELLIPSIS
    [<matplotlib.lines.Line2D object at 0x...>]
    """


    def __init__(self, plotter, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self._plotter = plotter


    def __call__(self, X, fig=None, **kwargs):
        """
        Plot the node using the specified plotting function

        Parameters
        ----------

        X : node

            The plotted node
        """
        kwargs_all = self._kwargs.copy()
        kwargs_all.update(kwargs)
        return self._plotter(X, *self._args, fig=fig, **kwargs_all)


class PDFPlotter(Plotter):
    r"""
    Plotter of probability density function of a scalar node

    Parameters
    ----------

    x_grid : array

        Numerical grid on which the density function is computed and
        plotted

    See also
    --------
    pdf
    """
    def __init__(self, x_grid, **kwargs):
        super().__init__(pdf, x_grid, **kwargs)


class ContourPlotter(Plotter):
    r"""
    Plotter of probability density function of a two-dimensional node

    Parameters
    ----------

    x1_grid : array

        Grid for the first dimension

    x2_grid : array

        Grid for the second dimension

    See also
    --------
    contour
    """
    def __init__(self, x1_grid, x2_grid, **kwargs):
        super().__init__(contour, x1_grid, x2_grid, **kwargs)


class HintonPlotter(Plotter):
    r"""
    Plotter of the Hinton diagram of a node

    See also
    --------
    hinton
    """
    def __init__(self, **kwargs):
        super().__init__(hinton, **kwargs)


class FunctionPlotter(Plotter):
    r"""
    Plotter of a node as a 1-dimensional function

    See also
    --------
    plot
    """
    def __init__(self, **kwargs):
        super().__init__(plot, **kwargs)


class GaussianMarkovChainPlotter(Plotter):
    r"""
    Plotter of a Gaussian Markov chain as a timeseries
    """
    def __init__(self, **kwargs):
        super().__init__(timeseries_gaussian_mc, **kwargs)


class GaussianTimeseriesPlotter(Plotter):
    r"""
    Plotter of a Gaussian node as a timeseries
    """
    def __init__(self, **kwargs):
        super().__init__(timeseries_gaussian, **kwargs)


class GaussianHintonPlotter(Plotter):
    r"""
    Plotter of a Gaussian node as a Hinton diagram
    """
    def __init__(self, **kwargs):
        super().__init__(gaussian_array, **kwargs)


class CategoricalMarkovChainPlotter(Plotter):
    r"""
    Plotter of a Categorical timeseries
    """
    def __init__(self, **kwargs):
        super().__init__(timeseries_categorical_mc, **kwargs)


def matrix_animation(A, filename=None, fps=25, fig=None, **kwargs):

    if fig is None:
        fig = plt.gcf()

    axes = fig.add_subplot(111)

    A = np.atleast_3d(A)
    vmax = np.max(np.abs(A))
    x = axes.imshow(A[0],
                    interpolation='nearest',
                    cmap='RdBu_r',
                    vmin=-vmax,
                    vmax=vmax,
                    **kwargs)
    s = axes.set_title('t = %d' % 0)

    def animate(nframe):
        s.set_text('t = %d' % nframe)
        x.set_array(A[nframe])
        return (x, s)

    anim = animation.FuncAnimation(fig, animate,
                                   frames=np.shape(A)[0],
                                   interval=1000/fps,
                                   blit=False,
                                   repeat=False)

    return anim


def save_animation(anim, filename, fps=25, bitrate=5000, fig=None):

    # A bug in numpy/matplotlib causes this not to work in python3.3:
    # https://github.com/matplotlib/matplotlib/issues/1891
    #
    # So the following command does not work currently..
    #
    # anim.save(filename, fps=fps)

    if fig is None:
        fig = plt.gcf()

    writer = animation.FFMpegFileWriter(fps=fps, bitrate=bitrate)
    writer.setup(fig, filename, 100)
    anim.save(filename,
              fps=fps,
              writer=writer,
              bitrate=bitrate)
    return


def binary_matrix(A, axes=None):
    if axes is None:
        axes = plt.gca()

    A = np.atleast_2d(A)
    G = np.zeros(np.shape(A) + (3,))
    G[A] = [0,0,0]
    G[np.logical_not(A)] = [1,1,1]
    axes.imshow(G, interpolation='nearest')


def gaussian_mixture_logpdf(x, w, mu, Sigma):
    # Shape(x)      = (N, D)
    # Shape(w)      = (K,)
    # Shape(mu)     = (K, D)
    # Shape(Sigma)  = (K, D, D)
    # Shape(result) = (N,)

    # Dimensionality
    D = np.shape(x)[-1]

    # Cholesky decomposition of the covariance matrix
    U = linalg.chol(Sigma)

    # Reshape x:
    # Shape(x)     = (N, 1, D)
    x = np.expand_dims(x, axis=-2)

    # (x-mu) and (x-mu)'*inv(Sigma)*(x-mu):
    # Shape(v)     = (N, K, D)
    # Shape(z)     = (N, K)
    v = x - mu
    z = np.einsum('...i,...i', v, linalg.chol_solve(U, v))

    # Log-determinant of Sigma:
    # Shape(ldet)  = (K,)
    ldet = linalg.chol_logdet(U)

    # Compute log pdf for each cluster:
    # Shape(lpdf)  = (N, K)
    lpdf = misc.gaussian_logpdf(z, 0, 0, ldet, D)



def matrixplot(A, colorbar=False, axes=None):
    if axes is None:
        axes = plt.gca()

    if sp.issparse(A):
        A = A.toarray()
    axes.imshow(A, interpolation='nearest')
    if colorbar:
        plt.colorbar(ax=axes)


def contourplot(x1, x2, y, colorbar=False, filled=True, axes=None):
    """ Plots 2D contour plot. x1 and x2 are 1D vectors, y contains
    the function values. y.size must be x1.size*x2.size. """

    if axes is None:
        axes = plt.gca()

    y = np.reshape(y, (len(x2),len(x1)))
    if filled:
        axes.contourf(x1, x2, y)
    else:
        axes.contour(x1, x2, y)
    if colorbar:
        plt.colorbar(ax=axes)


def errorplot(y=None, error=None, x=None, lower=None, upper=None,
              color=(0,0,0,1), fillcolor=None, axes=None, **kwargs):

    if axes is None:
        axes = plt.gca()

    # Default inputs
    if x is None:
        x = np.arange(np.size(y))

    # Parse errors (lower=lower/error/upper, upper=upper/error/lower)
    if lower is None:
        if error is not None:
            lower = error
        elif upper is not None:
            lower = upper
    if upper is None:
        if error is not None:
            upper = error
        elif lower is not None:
            upper = lower

    # Plot errors
    if (lower is not None) and (upper is not None):
        l = y - lower
        u = y + upper
        if fillcolor is None:
            color = colors.ColorConverter().to_rgba(color)
            fillcolor = tuple(color[:3]) + (0.2 * color[3],)
        axes.fill_between(x,
                          l,
                          u,
                          facecolor=fillcolor,
                          edgecolor=(0, 0, 0, 0),
                          linewidth=1,
                          interpolate=True)
    # Plot function
    axes.plot(x, y, color=color, **kwargs)


def plotmatrix(X):
    """
    Creates a matrix of marginal plots.

    On diagonal, are marginal plots of each variable. Off-diagonal plot (i,j)
    shows the joint marginal density of x_i and x_j.
    """
    return X.plotmatrix()


def _pdf_t(mu, s2, nu, axes=None, scale=4, color='k'):
    """
    """
    if axes is None:
        axes = plt.gca()

    s = np.sqrt(s2)
    x = np.linspace(mu-scale*s, mu+scale*s, num=100)
    y2 = (x-mu)**2 / s2
    lpdf = random.t_logpdf(y2, np.log(s2), nu, 1)
    p = np.exp(lpdf)
    return axes.plot(x, p, color=color)


def _pdf_gamma(a, b, axes=None, scale=4, color='k'):
    """
    """
    if axes is None:
        axes = plt.gca()

    if np.size(a) != 1 or np.size(b) != 1:
        raise ValueError("Parameters must be scalars")
    mean = a/b
    v = scale*np.sqrt(a/b**2)
    m = max(0, mean-v)
    n = mean + v
    x = np.linspace(m, n, num=100)
    logx = np.log(x)
    lpdf = random.gamma_logpdf(b*x,
                               logx,
                               a*logx,
                               a*np.log(b),
                               special.gammaln(a))
    p = np.exp(lpdf)
    return axes.plot(x, p, color=color)


def _contour_t(mu, Cov, nu, axes=None, scale=4, transpose=False, colors='k'):
    """
    """
    if axes is None:
        axes = plt.gca()

    if np.shape(mu) != (2,) or np.shape(Cov) != (2,2) or np.shape(nu) != ():
        print(np.shape(mu), np.shape(Cov), np.shape(nu))
        raise ValueError("Only 2-d t-distribution allowed")

    if transpose:
        mu = mu[[1,0]]
        Cov = Cov[np.ix_([1,0],[1,0])]

    s = np.sqrt(np.diag(Cov))
    x0 = np.linspace(mu[0]-scale*s[0], mu[0]+scale*s[0], num=100)
    x1 = np.linspace(mu[1]-scale*s[1], mu[1]+scale*s[1], num=100)
    X0X1 = misc.grid(x0, x1)
    Y = X0X1 - mu
    L = linalg.chol(Cov)
    logdet_Cov = linalg.chol_logdet(L)
    Z = linalg.chol_solve(L, Y)
    Z = linalg.inner(Y, Z, ndim=1)
    lpdf = random.t_logpdf(Z, logdet_Cov, nu, 2)
    p = np.exp(lpdf)
    shape = (np.size(x0), np.size(x1))
    X0 = np.reshape(X0X1[:,0], shape)
    X1 = np.reshape(X0X1[:,1], shape)
    P = np.reshape(p, shape)
    return axes.contour(X0, X1, P, colors=colors)


def _contour_gaussian_gamma(mu, s2, a, b, axes=None, transpose=False):
    """
    """
    pass


def ellipse_from_cov(xy, cov, scale=2, **kwargs):
    """
    Create an ellipse from a covariance matrix.

    Parameters
    ----------
    xy : np.ndarray
        position of the ellipse
    cov : np.ndarray
        covariance matrix
    scale : float
        scale of the ellipse (default is three standard deviations)
    kwargs : dict
        keyword arguments passed on to `matplotlib.patches.Ellipse`

    Returns
    -------
    ellipse : matplotlib.patches.Ellipse
    """
    evals, evecs = np.linalg.eigh(cov)
    angle = np.arctan2(*evecs[::-1, 0])
    width, height = scale * np.sqrt(evals)
    return patches.Ellipse(xy, width, height, np.rad2deg(angle), **kwargs)


def ellipse_from_precision(xy, precision, scale=2, **kwargs):
    """
    Create an ellipse from a covariance matrix.

    Parameters
    ----------
    xy : np.ndarray
        position of the ellipse
    cov : np.ndarray
        covariance matrix
    scale : float
        scale of the ellipse (default is three standard deviations)
    kwargs : dict
        keyword arguments passed on to `matplotlib.patches.Ellipse`

    Returns
    -------
    ellipse : matplotlib.patches.Ellipse
    """
    return ellipse_from_cov(xy, np.linalg.inv(precision), scale, **kwargs)



################################################################################
# Copyright (C) 2013 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


"""
Unit tests for `deterministic` module.
"""

import unittest

import numpy as np
import scipy

from numpy import testing

from ..node import Node, Moments
from ..deterministic import tile
from ..stochastic import Stochastic


class TestTile(unittest.TestCase):

    def check_message_to_children(self, tiles, u_parent, u_tiled,
                                  dims=None, plates=None):
        # Set up the dummy model
        class Dummy(Stochastic):
            _parent_moments = ()
            _moments = Moments()
            def __init__(self, u, dims=dims, plates=plates):
                super().__init__(dims=dims, plates=plates, initialize=False)
                self.u = u

        X = Dummy(u_parent, dims=dims, plates=plates)
        Y = tile(X, tiles)

        u_Y = Y._compute_moments(u_parent)

        for (x,y) in zip(u_Y, u_tiled):
            self.assertEqual(np.shape(x), np.shape(y),
                             msg="Incorrect shape.")
            testing.assert_allclose(x, y,
                                    err_msg="Incorrect moments.")


    def test_message_to_children(self):
        """
        Test the moments of Tile node.
        """
        # Define th check function
        check_message_to_children = self.check_message_to_children
        # Check scalar (and broadcasting)
        check_message_to_children(2,
                                  (5,),
                                  (5,), 
                                  dims=[()],
                                  plates=()),
        # Check 1-D
        check_message_to_children(2,
                                  ([1,2],),
                                  ([1,2,1,2],),
                                  dims=[()],
                                  plates=(2,))
        # Check N-D
        check_message_to_children(2,
                                  ([[1,2],
                                    [3,4],
                                    [5,6]],),
                                  ([[1,2,1,2],
                                    [3,4,3,4],
                                    [5,6,5,6]],),
                                  dims=[()],
                                  plates=(3,2))
        # Check not-last plate
        check_message_to_children([2,1],
                                  ([[1,2],
                                    [3,4]],),
                                  ([[1,2],
                                    [3,4],
                                    [1,2],
                                    [3,4]],),
                                  dims=[()],
                                  plates=(2,2))
        # Check several plates
        check_message_to_children([2,3],
                                  ([[1,2],
                                    [3,4]],),
                                  ([[1,2,1,2,1,2],
                                    [3,4,3,4,3,4],
                                    [1,2,1,2,1,2],
                                    [3,4,3,4,3,4]],),
                                  dims=[()],
                                  plates=(2,2))
        # Check non-zero dimensional variables
        check_message_to_children(2,
                                  ([[1,2],
                                    [3,4]],),
                                  ([[1,2],
                                    [3,4],
                                    [1,2],
                                    [3,4]],),
                                  dims=[(2,)],
                                  plates=(2,))
        # Check several moments
        check_message_to_children(2,
                                  ([[1,2],
                                    [3,4]],
                                   [1,2]),
                                  ([[1,2],
                                    [3,4],
                                    [1,2],
                                    [3,4]],
                                   [1,2,1,2]),
                                  dims=[(2,),()],
                                  plates=(2,))
        # Check broadcasting of tiled plate
        check_message_to_children(2,
                                  ([[1,],
                                    [2,]],),
                                  ([[1,],
                                    [2,]],),
                                  dims=[()],
                                  plates=(2,2))
        # Check broadcasting of non-tiled plate
        check_message_to_children(2,
                                  ([[1,2]],),
                                  ([[1,2,1,2]],),
                                  dims=[()],
                                  plates=(2,2))
        # Check broadcasting of leading plates that are not in parent
        check_message_to_children([2,1],
                                  ([1,2],),
                                  ([1,2],),
                                  dims=[()],
                                  plates=(2,))
        
        
    def check_message_to_parent(self, tiles, m_children, m_true,
                                dims=None, plates_parent=None,
                                plates_children=None):
        # Set up the dummy model
        class Dummy(Stochastic):
            _parent_moments = ()
            _moments = Moments()
        X = Dummy(dims=dims, plates=plates_parent, initialize=False)
        Y = tile(X, tiles)

        m = Y._compute_message_to_parent(0, m_children, None)

        for (x,y) in zip(m, m_true):
            self.assertEqual(np.shape(x), np.shape(y),
                             msg="Incorrect shape.")
            testing.assert_allclose(x, y,
                                    err_msg="Incorrect message.")


    def test_message_to_parent(self):
        """
        Test the parent message of Tile node.
        """
        # Define th check function
        check = self.check_message_to_parent
        # Check scalar
        check(2,
              ([5,5],), 
              (10,),
              dims=[()],
              plates_parent=(),
              plates_children=(2,)),
        # Check 1-D
        check(2,
              ([1,2,3,4],),
              ([4,6],),
              dims=[()],
              plates_parent=(2,),
              plates_children=(4,))
        # Check N-D
        check(2,
              ([[1,2,7,8],
                [3,4,9,0],
                [5,6,1,2]],),
              ([[8,10],
                [12,4],
                [6,8]],),
              dims=[()],
              plates_parent=(3,2),
              plates_children=(3,4))
        # Check not-last plate
        check([2,1],
              ([[1,2],
                [3,4],
                [5,6],
                [7,8]],),
              ([[6,8],
                [10,12]],),
              dims=[()],
              plates_parent=(2,2),
              plates_children=(4,2))
        # Check several plates
        check([2,3],
              ([[1,2,1,2,1,2],
                [3,4,3,4,3,4],
                [1,2,1,2,1,2],
                [3,4,3,4,3,4]],),
              ([[6,12],
                [18,24]],),
              dims=[()],
              plates_parent=(2,2),
              plates_children=(4,6))
        # Check broadcasting if message has unit axis for tiled plate
        check(2,
              ([[1,],
                [2,],
                [3,]],),
              ([[2,],
                [4,],
                [6,]],),
              dims=[()],
              plates_parent=(3,2),
              plates_children=(3,4))
        # Check broadcasting if message has unit axis for non-tiled plate
        check(2,
              ([[1,2,3,4]],),
              ([[4,6]],),
              dims=[()],
              plates_parent=(3,2),
              plates_children=(3,4))
        # Check non-zero dimensional variables
        check(2,
              ([[1,2],
                [3,4],
                [5,6],
                [7,8]],),
              ([[6,8],
                [10,12]],),
              dims=[(2,)],
              plates_parent=(2,),
              plates_children=(4,))
        # Check several moments
        check(2,
              ([[1,2],
                [3,4],
                [5,6],
                [7,8]],
               [1,2,3,4]),
              ([[6,8],
                [10,12]],
               [4,6]),
              dims=[(2,),()],
              plates_parent=(2,),
              plates_children=(4,))
        
    def check_mask_to_parent(self, tiles, mask_child, mask_true,
                             plates_parent=None,
                             plates_children=None):
        # Set up the dummy model
        class Dummy(Stochastic):
            _moments = Moments()
            _parent_moments = ()
        X = Dummy(dims=[()], plates=plates_parent, initialize=False)
        Y = tile(X, tiles)

        mask = Y._compute_weights_to_parent(0, mask_child) != 0

        self.assertEqual(np.shape(mask), np.shape(mask_true),
                         msg="Incorrect shape.")
        testing.assert_equal(mask, mask_true,
                             err_msg="Incorrect mask.")


    def test_mask_to_parent(self):
        """
        Test the mask message to parent of Tile node.
        """
        # Define th check function
        check = self.check_mask_to_parent
        # Check scalar parent
        check(2,
              [True,False], 
              True,
              plates_parent=(),
              plates_children=(2,))
        check(2,
              [False,False], 
              False,
              plates_parent=(),
              plates_children=(2,))
        # Check 1-D
        check(2,
              [True,False,False,False],
              [True,False],
              plates_parent=(2,),
              plates_children=(4,))
        # Check N-D
        check(2,
              [[True,False,True,False],
               [False,True,False,False]],
              [[True,False],
               [False,True]],
              plates_parent=(2,2),
              plates_children=(2,4))
        # Check not-last plate
        check([2,1],
              [[True,False],
               [False,True],
               [True,False],
               [False,False]],
              [[True,False],
               [False,True]],
              plates_parent=(2,2),
              plates_children=(4,2))
        # Check several plates
        check([2,3],
              [[False,False,True,False,False,False],
               [False,False,False,False,False,False],
               [False,False,False,False,False,False],
               [False,True,False,False,False,False]],
              [[True,False],
               [False,True]],
              plates_parent=(2,2),
              plates_children=(4,6))
        # Check broadcasting if message has unit axis for tiled plate
        check(2,
              [[True,],
               [False,],
               [True,]],
              [[True,],
               [False,],
               [True,]],
              plates_parent=(3,2),
              plates_children=(3,4))
        # Check broadcasting if message has unit axis for non-tiled plate
        check(2,
              [[False,False,False,True]],
              [[False,True]],
              plates_parent=(3,2),
              plates_children=(3,4))
        
        
        

